from typing import (
    List,
    Optional,
    Tuple,
    Union,
)

from lxml import etree
from lxml.etree import _Element

from pcs import settings
from pcs.common.tools import xml_fromstring
from pcs.lib.external import CommandRunner

from . import const
from .error import UnableToGetAgentMetadata
from .types import (
    FakeAgentName,
    ResourceAgentActionOcf1_0,
    ResourceAgentActionOcf1_1,
    ResourceAgentMetadataOcf1_0,
    ResourceAgentMetadataOcf1_1,
    ResourceAgentName,
    ResourceAgentParameterOcf1_0,
    ResourceAgentParameterOcf1_1,
)

### load metadata


def _load_metadata_xml(
    runner: CommandRunner, agent_name: ResourceAgentName
) -> str:
    """
    Run pacemaker tool to get raw metadata from an agent

    runner -- external processes runner
    agent_name -- name of an agent whose metadata we want to get
    """
    env_path = ":".join(
        [
            # otherwise pacemaker cannot run RHEL fence agents to get their
            # metadata
            settings.fence_agent_binaries,
            # otherwise heartbeat and cluster-glue agents don't work
            "/bin",
            # otherwise heartbeat and cluster-glue agents don't work
            "/usr/bin",
        ]
    )
    stdout, stderr, retval = runner.run(
        [settings.crm_resource_binary, "--show-metadata", agent_name.full_name],
        env_extend={"PATH": env_path},
    )
    if retval != 0:
        raise UnableToGetAgentMetadata(agent_name.full_name, stderr.strip())
    return stdout.strip()


def _load_fake_agent_metadata_xml(
    runner: CommandRunner, agent_name: FakeAgentName
) -> str:
    """
    Run pacemaker tool to get raw metadata from pacemaker

    runner -- external processes runner
    agent_name -- name of pacemaker part whose metadata we want to get
    """
    name_to_executable = {
        const.PACEMAKER_FENCED: settings.pacemaker_fenced,
    }
    if agent_name not in name_to_executable:
        raise UnableToGetAgentMetadata(agent_name, "Unknown agent")
    stdout, stderr, dummy_retval = runner.run(
        [name_to_executable[agent_name], "metadata"]
    )
    metadata = stdout.strip()
    if not metadata:
        raise UnableToGetAgentMetadata(agent_name, stderr.strip())
    return metadata


def _get_ocf_version(metadata: _Element) -> str:
    """
    Extract OCF version from agent metadata XML

    metadata -- metadata XML document
    """
    version = metadata.findtext("./version")
    return const.OCF_1_0 if version is None else version.strip()


def _metadata_xml_to_dom(metadata: str) -> _Element:
    """
    Parse metadata string to XML document and validate against RNG schema

    metadata -- metadata string to parse
    """
    dom = xml_fromstring(metadata)
    ocf_version = _get_ocf_version(dom)
    if ocf_version == const.OCF_1_1:
        etree.RelaxNG(file=settings.path.ocf_1_1_schema).assertValid(dom)
    return dom


def load_metadata(
    runner: CommandRunner, agent_name: ResourceAgentName
) -> _Element:
    """
    Return specified agent's metadata as an XML document

    runner -- external processes runner
    agent_name -- name of an agent whose metadata we want to get
    """
    try:
        return _metadata_xml_to_dom(_load_metadata_xml(runner, agent_name))
    except (etree.XMLSyntaxError, etree.DocumentInvalid) as e:
        raise UnableToGetAgentMetadata(agent_name.full_name, str(e)) from e


def load_fake_agent_metadata(
    runner: CommandRunner, agent_name: FakeAgentName
) -> _Element:
    """
    Return pacemaker metadata as an XML document

    runner -- external processes runner
    agent_name -- name of pacemaker part whose metadata we want to get
    """
    try:
        return _metadata_xml_to_dom(
            _load_fake_agent_metadata_xml(runner, agent_name)
        )
    except (etree.XMLSyntaxError, etree.DocumentInvalid) as e:
        raise UnableToGetAgentMetadata(agent_name, str(e)) from e


### parse metadata


def parse_metadata(
    name: ResourceAgentName,
    metadata: _Element,
) -> Tuple[
    Union[ResourceAgentMetadataOcf1_0, ResourceAgentMetadataOcf1_1], str
]:
    """
    Parse XML metadata to a dataclass

    name -- name of an agent
    metadata -- metadata XML document
    """
    ocf_version = _get_ocf_version(metadata)
    if ocf_version == const.OCF_1_1:
        return _parse_agent_1_1(name, metadata), ocf_version
    return _parse_agent_1_0(name, metadata), ocf_version


def _parse_agent_1_0(
    name: ResourceAgentName, metadata: _Element
) -> ResourceAgentMetadataOcf1_0:
    """
    Parse OCF 1.0 XML metadata to a dataclass

    name -- name of an agent
    metadata -- metadata XML document
    """
    parameters_el = metadata.find("./parameters")
    actions_el = metadata.find("./actions")
    return ResourceAgentMetadataOcf1_0(
        name=name,
        shortdesc=_get_shortdesc(metadata) or metadata.get("shortdesc"),
        longdesc=_get_longdesc(metadata),
        parameters=(
            []
            if parameters_el is None
            else _parse_parameters_1_0(parameters_el)
        ),
        actions=([] if actions_el is None else _parse_actions_1_0(actions_el)),
    )


def _parse_agent_1_1(
    name: ResourceAgentName, metadata: _Element
) -> ResourceAgentMetadataOcf1_1:
    """
    Parse OCF 1.1 XML metadata to a dataclass

    name -- name of an agent
    metadata -- metadata XML document
    """
    parameters_el = metadata.find("./parameters")
    actions_el = metadata.find("./actions")
    return ResourceAgentMetadataOcf1_1(
        name=name,
        shortdesc=_get_shortdesc(metadata),
        longdesc=_get_longdesc(metadata),
        parameters=(
            []
            if parameters_el is None
            else _parse_parameters_1_1(parameters_el)
        ),
        actions=([] if actions_el is None else _parse_actions_1_1(actions_el)),
    )


def _parse_parameter_content(
    parameter_el: _Element,
) -> Tuple[str, Optional[str], Optional[List[str]]]:
    value_type = "string"
    default_value = None
    enum_values = None
    content_el = parameter_el.find("content")
    if content_el is not None:
        value_type = content_el.get("type", value_type)
        default_value = content_el.get("default", default_value)
        if value_type == "select":
            enum_values = [
                str(option_el.attrib["value"])
                for option_el in content_el.iterfind("./option")
            ]
    return value_type, default_value, enum_values


def _parse_parameters_1_0(
    element: _Element,
) -> List[ResourceAgentParameterOcf1_0]:
    result = []
    for parameter_el in element.iter("parameter"):
        value_type, default_value, enum_values = _parse_parameter_content(
            parameter_el
        )
        result.append(
            ResourceAgentParameterOcf1_0(
                name=str(parameter_el.get("name", "")),
                shortdesc=_get_shortdesc(parameter_el),
                longdesc=_get_longdesc(parameter_el),
                type=value_type,
                default=default_value,
                enum_values=enum_values,
                required=parameter_el.get("required"),
                deprecated=parameter_el.get("deprecated"),
                obsoletes=parameter_el.get("obsoletes"),
                unique=parameter_el.get("unique"),
            )
        )
    return result


def _parse_parameters_1_1(
    element: _Element,
) -> List[ResourceAgentParameterOcf1_1]:
    result = []
    for parameter_el in element.iter("parameter"):
        value_type, default_value, enum_values = _parse_parameter_content(
            parameter_el
        )

        deprecated, deprecated_by, deprecated_desc = False, [], None
        deprecated_el = parameter_el.find("deprecated")
        if deprecated_el is not None:
            deprecated = True
            deprecated_by = [
                str(replaced_with_el.attrib["name"])
                for replaced_with_el in deprecated_el.iterfind("replaced-with")
            ]
            deprecated_desc = _get_desc(deprecated_el)

        result.append(
            ResourceAgentParameterOcf1_1(
                name=str(parameter_el.attrib["name"]),
                shortdesc=_get_shortdesc(parameter_el),
                longdesc=_get_longdesc(parameter_el),
                type=value_type,
                default=default_value,
                enum_values=enum_values,
                required=parameter_el.get("required"),
                deprecated=deprecated,
                deprecated_by=deprecated_by,
                deprecated_desc=deprecated_desc,
                unique_group=parameter_el.get("unique-group"),
                reloadable=parameter_el.get("reloadable"),
            )
        )
    return result


def _parse_actions_1_0(element: _Element) -> List[ResourceAgentActionOcf1_0]:
    return [
        ResourceAgentActionOcf1_0(
            name=str(action.get("name", "")),
            timeout=action.get("timeout"),
            interval=action.get("interval"),
            role=action.get("role"),
            start_delay=action.get("start-delay"),
            depth=action.get("depth"),
            automatic=action.get("automatic"),
            on_target=action.get("on_target"),
        )
        for action in element.iter("action")
    ]


def _parse_actions_1_1(element: _Element) -> List[ResourceAgentActionOcf1_1]:
    return [
        ResourceAgentActionOcf1_1(
            name=str(action.attrib["name"]),
            timeout=action.get("timeout"),
            interval=action.get("interval"),
            role=action.get("role"),
            start_delay=action.get("start-delay"),
            depth=action.get("depth"),
            automatic=action.get("automatic"),
            on_target=action.get("on_target"),
        )
        for action in element.iter("action")
    ]


def _get_desc(element: _Element) -> Optional[str]:
    return _get_text_from_dom_element(element.find("desc"))


def _get_shortdesc(element: _Element) -> Optional[str]:
    return _get_text_from_dom_element(element.find("shortdesc"))


def _get_longdesc(element: _Element) -> Optional[str]:
    return _get_text_from_dom_element(element.find("longdesc"))


def _get_text_from_dom_element(element: Optional[_Element]) -> Optional[str]:
    return (
        None
        if element is None or element.text is None
        else str(element.text).strip()
    )



import json

from pcs.common import reports
from pcs.common.node_communicator import (
    Request,
    RequestData,
)
from pcs.common.reports import ReportItemSeverity
from pcs.common.reports.item import ReportItem
from pcs.lib.communication.tools import (
    AllAtOnceStrategyMixin,
    AllSameDataMixin,
    MarkSuccessfulMixin,
    OneByOneStrategyMixin,
    RunRemotelyBase,
    SimpleResponseProcessingMixin,
)
from pcs.lib.node_communication import response_to_report_item
from pcs.lib.tools import environment_file_to_dict


class ServiceAction(
    SimpleResponseProcessingMixin,
    AllSameDataMixin,
    AllAtOnceStrategyMixin,
    RunRemotelyBase,
):
    def _get_request_action(self):
        raise NotImplementedError()

    def _get_before_report(self):
        raise NotImplementedError()

    def _get_success_report(self, node_label):
        raise NotImplementedError()

    def _get_request_data(self):
        return RequestData(self._get_request_action())

    def before(self):
        self._report(self._get_before_report())


class EnableSbdService(ServiceAction):
    def _get_request_action(self):
        return "remote/sbd_enable"

    def _get_before_report(self):
        return ReportItem.info(
            reports.messages.ServiceActionStarted(
                reports.const.SERVICE_ACTION_ENABLE, "sbd"
            )
        )

    def _get_success_report(self, node_label):
        return ReportItem.info(
            reports.messages.ServiceActionSucceeded(
                reports.const.SERVICE_ACTION_ENABLE, "sbd", node_label
            )
        )


class DisableSbdService(ServiceAction):
    def _get_request_action(self):
        return "remote/sbd_disable"

    def _get_before_report(self):
        return ReportItem.info(
            reports.messages.ServiceActionStarted(
                reports.const.SERVICE_ACTION_DISABLE,
                "sbd",
            )
        )

    def _get_success_report(self, node_label):
        return ReportItem.info(
            reports.messages.ServiceActionSucceeded(
                reports.const.SERVICE_ACTION_DISABLE, "sbd", node_label
            )
        )


class StonithWatchdogTimeoutAction(
    AllSameDataMixin,
    MarkSuccessfulMixin,
    OneByOneStrategyMixin,
    RunRemotelyBase,
):
    def _get_request_action(self):
        raise NotImplementedError()

    def _get_request_data(self):
        return RequestData(self._get_request_action())

    def _process_response(self, response):
        report_item = response_to_report_item(
            response, severity=ReportItemSeverity.WARNING
        )
        if report_item is None:
            self._on_success()
            return []
        self._report(report_item)
        return self._get_next_list()


class RemoveStonithWatchdogTimeout(StonithWatchdogTimeoutAction):
    def _get_request_action(self):
        return "remote/remove_stonith_watchdog_timeout"


class SetStonithWatchdogTimeoutToZero(StonithWatchdogTimeoutAction):
    def _get_request_action(self):
        return "remote/set_stonith_watchdog_timeout_to_zero"


class SetSbdConfig(
    SimpleResponseProcessingMixin, AllAtOnceStrategyMixin, RunRemotelyBase
):
    def __init__(self, report_processor):
        super().__init__(report_processor)
        self._request_data_list = []

    def _prepare_initial_requests(self):
        return [
            Request(
                target,
                RequestData("remote/set_sbd_config", [("config", config)]),
            )
            for target, config in self._request_data_list
        ]

    def _get_success_report(self, node_label):
        return ReportItem.info(
            reports.messages.SbdConfigAcceptedByNode(node_label)
        )

    def add_request(self, target, config):
        self._request_data_list.append((target, config))

    def before(self):
        self._report(
            ReportItem.info(reports.messages.SbdConfigDistributionStarted())
        )


class GetSbdConfig(AllSameDataMixin, AllAtOnceStrategyMixin, RunRemotelyBase):
    def __init__(self, report_processor):
        super().__init__(report_processor)
        self._config_list = []
        self._successful_target_list = []

    def _get_request_data(self):
        return RequestData("remote/get_sbd_config")

    def _process_response(self, response):
        report_item = response_to_report_item(
            response, severity=ReportItemSeverity.WARNING
        )
        node_label = response.request.target.label
        if report_item is not None:
            if not response.was_connected:
                self._report(report_item)
            self._report(
                ReportItem.warning(
                    reports.messages.UnableToGetSbdConfig(node_label, "")
                )
            )
            return
        self._config_list.append(
            {
                "node": node_label,
                "config": environment_file_to_dict(response.data),
            }
        )
        self._successful_target_list.append(node_label)

    def on_complete(self):
        for node in self._target_list:
            if node.label not in self._successful_target_list:
                self._config_list.append({"node": node.label, "config": None})
        return self._config_list


class GetSbdStatus(AllSameDataMixin, AllAtOnceStrategyMixin, RunRemotelyBase):
    def __init__(self, report_processor):
        super().__init__(report_processor)
        self._status_list = []
        self._successful_target_list = []

    def _get_request_data(self):
        return RequestData(
            "remote/check_sbd",
            # here we just need info about sbd service, therefore watchdog and
            # device list is empty
            [
                ("watchdog", ""),
                ("device_list", "[]"),
            ],
        )

    def _process_response(self, response):
        report_item = response_to_report_item(
            response, severity=ReportItemSeverity.WARNING
        )
        node_label = response.request.target.label
        if report_item is not None:
            self._report_list(
                [
                    report_item,
                    # reason is in previous report item, warning is there
                    # implicit
                    ReportItem.warning(
                        reports.messages.UnableToGetSbdStatus(node_label, "")
                    ),
                ]
            )
            return
        try:
            self._status_list.append(
                {"node": node_label, "status": json.loads(response.data)["sbd"]}
            )
            self._successful_target_list.append(node_label)
        except (ValueError, KeyError) as e:
            self._report(
                ReportItem.warning(
                    reports.messages.UnableToGetSbdStatus(node_label, str(e))
                )
            )

    def on_complete(self):
        for node in self._target_list:
            if node.label not in self._successful_target_list:
                self._status_list.append(
                    {
                        "node": node.label,
                        "status": {
                            "installed": None,
                            "enabled": None,
                            "running": None,
                        },
                    }
                )
        return self._status_list


class CheckSbd(AllAtOnceStrategyMixin, RunRemotelyBase):
    def __init__(self, report_processor):
        super().__init__(report_processor)
        self._request_data_list = []

    def _prepare_initial_requests(self):
        return [
            Request(
                target,
                RequestData(
                    "remote/check_sbd",
                    [
                        ("watchdog", watchdog),
                        ("device_list", json.dumps(device_list)),
                    ],
                ),
            )
            for target, watchdog, device_list in self._request_data_list
        ]

    def _process_response(self, response):
        report_item = response_to_report_item(response)
        if report_item:
            self._report(report_item)
            return
        report_list = []
        node_label = response.request.target.label
        try:
            data = json.loads(response.data)
            if not data["sbd"]["installed"]:
                report_list.append(
                    ReportItem.error(
                        reports.messages.SbdNotInstalled(node_label)
                    )
                )
            if "watchdog" in data:
                if data["watchdog"]["exist"]:
                    if not data["watchdog"].get("is_supported", True):
                        report_list.append(
                            ReportItem.error(
                                reports.messages.SbdWatchdogNotSupported(
                                    node_label, data["watchdog"]["path"]
                                )
                            )
                        )
                else:
                    report_list.append(
                        ReportItem.error(
                            reports.messages.WatchdogNotFound(
                                node_label, data["watchdog"]["path"]
                            )
                        )
                    )

            for device in data.get("device_list", []):
                if not device["exist"]:
                    report_list.append(
                        ReportItem.error(
                            reports.messages.SbdDeviceDoesNotExist(
                                device["path"], node_label
                            )
                        )
                    )
                elif not device["block_device"]:
                    report_list.append(
                        ReportItem.error(
                            reports.messages.SbdDeviceIsNotBlockDevice(
                                device["path"], node_label
                            )
                        )
                    )
                # TODO maybe we can check whenever device is initialized by sbd
                # (by running 'sbd -d <dev> dump;')
        except (ValueError, KeyError, TypeError):
            report_list.append(
                ReportItem.error(
                    reports.messages.InvalidResponseFormat(node_label)
                )
            )
        if report_list:
            self._report_list(report_list)
        else:
            self._report(
                ReportItem.info(
                    reports.messages.SbdCheckSuccess(
                        response.request.target.label
                    )
                )
            )

    def add_request(self, target, watchdog, device_list):
        self._request_data_list.append((target, watchdog, device_list))

    def before(self):
        self._report(ReportItem.info(reports.messages.SbdCheckStarted()))



import unittest

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.misc import ParametrizedTestMetaClass
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    outdent,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner

empty_cib = rc("cib-empty.xml")

ERRORS_HAVE_OCCURRED = (
    "Error: Errors have occurred, therefore pcs is unable to continue\n"
)


class PcsAlertTest(unittest.TestCase, AssertPcsMixin):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_alert")
        write_file_to_tmpfile(empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()


class CreateAlertTest(PcsAlertTest):
    def test_create_multiple_without_id(self):
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 No alerts defined
""",
        )

        self.assert_pcs_success("alert create path=test".split())
        self.assert_pcs_success("alert create path=test".split())
        self.assert_pcs_success("alert create path=test2".split())
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert (path=test)
 Alert: alert-1 (path=test)
 Alert: alert-2 (path=test2)
""",
        )

    def test_create_multiple_with_id(self):
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 No alerts defined
""",
        )
        self.assert_pcs_success("alert create id=alert1 path=test".split())
        self.assert_pcs_success(
            "alert create id=alert2 description=desc path=test".split()
        )
        self.assert_pcs_success(
            "alert create description=desc2 path=test2 id=alert3".split()
        )
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert1 (path=test)
 Alert: alert2 (path=test)
  Description: desc
 Alert: alert3 (path=test2)
  Description: desc2
""",
        )

    def test_create_with_options(self):
        self.assert_pcs_success(
            (
                "alert create id=alert1 description=desc path=test "
                "options opt2=val2 opt1=val1 meta m2=v2 m1=v1"
            ).split()
        )
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert1 (path=test)
  Description: desc
  Options: opt1=val1 opt2=val2
  Meta options: m1=v1 m2=v2
""",
        )

    def test_already_exists(self):
        self.assert_pcs_success("alert create id=alert1 path=test".split())
        self.assert_pcs_fail(
            "alert create id=alert1 path=test".split(),
            "Error: 'alert1' already exists\n",
        )
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert1 (path=test)
""",
        )

    def test_path_is_required(self):
        self.assert_pcs_fail(
            "alert create id=alert1".split(),
            "Error: required option 'path' is missing\n",
        )


class UpdateAlertTest(PcsAlertTest):
    def test_update_everything(self):
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 No alerts defined
""",
        )
        self.assert_pcs_success(
            (
                "alert create id=alert1 description=desc path=test "
                "options opt1=val1 opt3=val3 meta m1=v1 m3=v3"
            ).split()
        )
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert1 (path=test)
  Description: desc
  Options: opt1=val1 opt3=val3
  Meta options: m1=v1 m3=v3
""",
        )
        self.assert_pcs_success(
            (
                "alert update alert1 description=new_desc path=/new/path "
                "options opt1= opt2=test opt3=1 meta m1= m2=v m3=3"
            ).split()
        )
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert1 (path=/new/path)
  Description: new_desc
  Options: opt2=test opt3=1
  Meta options: m2=v m3=3
""",
        )

    def test_not_existing_alert(self):
        self.assert_pcs_fail(
            "alert update alert1".split(),
            "Error: alert 'alert1' does not exist\n",
        )


class DeleteRemoveAlertTest(PcsAlertTest):
    command = None

    def _test_usage(self):
        self.assert_pcs_fail(
            ["alert", self.command],
            stdout_start=f"\nUsage: pcs alert <command>\n    {self.command} <",
        )

    def _test_not_existing_alert(self):
        self.assert_pcs_fail(
            ["alert", self.command, "alert1"],
            ("Error: alert 'alert1' does not exist\n" + ERRORS_HAVE_OCCURRED),
        )

    def _test_one(self):
        self.assert_pcs_success(
            "alert config".split(),
            outdent(
                """\
                Alerts:
                 No alerts defined
                """
            ),
        )

        self.assert_pcs_success("alert create path=test id=alert1".split())
        self.assert_pcs_success(
            "alert config".split(),
            outdent(
                """\
                Alerts:
                 Alert: alert1 (path=test)
                """
            ),
        )
        self.assert_pcs_success(["alert", self.command, "alert1"])
        self.assert_pcs_success(
            "alert config".split(),
            outdent(
                """\
                Alerts:
                 No alerts defined
                """
            ),
        )

    def _test_multiple(self):
        self.assert_pcs_success(
            "alert config".split(),
            outdent(
                """\
                Alerts:
                 No alerts defined
                """
            ),
        )

        self.assert_pcs_success("alert create path=test id=alert1".split())
        self.assert_pcs_success("alert create path=test id=alert2".split())
        self.assert_pcs_success("alert create path=test id=alert3".split())
        self.assert_pcs_success(
            "alert config".split(),
            outdent(
                """\
                Alerts:
                 Alert: alert1 (path=test)
                 Alert: alert2 (path=test)
                 Alert: alert3 (path=test)
                """
            ),
        )
        self.assert_pcs_success(["alert", self.command, "alert1", "alert3"])
        self.assert_pcs_success(
            "alert config".split(),
            outdent(
                """\
                Alerts:
                 Alert: alert2 (path=test)
                """
            ),
        )


class DeleteAlertTest(
    DeleteRemoveAlertTest, metaclass=ParametrizedTestMetaClass
):
    command = "delete"


class RemoveAlertTest(
    DeleteRemoveAlertTest, metaclass=ParametrizedTestMetaClass
):
    command = "remove"


class AddRecipientTest(PcsAlertTest):
    def test_success(self):
        self.assert_pcs_success("alert create path=test".split())
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert (path=test)
""",
        )
        self.assert_pcs_success(
            "alert recipient add alert value=rec_value".split()
        )
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=rec_value)
""",
        )
        self.assert_pcs_success(
            (
                "alert recipient add alert value=rec_value2 id=my-recipient "
                "description=description options o2=2 o1=1 meta m2=v2 m1=v1"
            ).split()
        )
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=rec_value)
   Recipient: my-recipient (value=rec_value2)
    Description: description
    Options: o1=1 o2=2
    Meta options: m1=v1 m2=v2
""",
        )

    def test_already_exists(self):
        self.assert_pcs_success("alert create path=test".split())
        self.assert_pcs_success(
            "alert recipient add alert value=rec_value id=rec".split()
        )
        self.assert_pcs_fail(
            "alert recipient add alert value=value id=rec".split(),
            "Error: 'rec' already exists\n",
        )
        self.assert_pcs_fail(
            "alert recipient add alert value=value id=alert".split(),
            "Error: 'alert' already exists\n",
        )

    def test_same_value(self):
        self.assert_pcs_success("alert create path=test".split())
        self.assert_pcs_success(
            "alert recipient add alert value=rec_value id=rec".split()
        )
        self.assert_pcs_fail(
            "alert recipient add alert value=rec_value".split(),
            (
                "Error: Recipient 'rec_value' in alert 'alert' already exists, "
                "use --force to override\n" + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: rec (value=rec_value)
""",
        )
        self.assert_pcs_success(
            "alert recipient add alert value=rec_value --force".split(),
            "Warning: Recipient 'rec_value' in alert 'alert' already exists\n",
        )
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: rec (value=rec_value)
   Recipient: alert-recipient (value=rec_value)
""",
        )

    def test_no_value(self):
        self.assert_pcs_success("alert create path=test".split())
        self.assert_pcs_fail(
            "alert recipient add alert id=rec".split(),
            "Error: required option 'value' is missing\n",
        )


class UpdateRecipientAlert(PcsAlertTest):
    def test_success(self):
        self.assert_pcs_success("alert create path=test".split())
        self.assert_pcs_success(
            (
                "alert recipient add alert value=rec_value "
                "description=description options o1=1 o3=3 meta m1=v1 m3=v3"
            ).split()
        )
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=rec_value)
    Description: description
    Options: o1=1 o3=3
    Meta options: m1=v1 m3=v3
""",
        )
        self.assert_pcs_success(
            (
                "alert recipient update alert-recipient value=new "
                "description=desc options o1= o2=v2 o3=3 meta m1= m2=2 m3=3"
            ).split()
        )
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=new)
    Description: desc
    Options: o2=v2 o3=3
    Meta options: m2=2 m3=3
""",
        )
        self.assert_pcs_success(
            "alert recipient update alert-recipient value=new".split()
        )
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=new)
    Description: desc
    Options: o2=v2 o3=3
    Meta options: m2=2 m3=3
""",
        )

    def test_value_exists(self):
        self.assert_pcs_success("alert create path=test".split())
        self.assert_pcs_success(
            "alert recipient add alert value=rec_value".split()
        )
        self.assert_pcs_success("alert recipient add alert value=value".split())
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=rec_value)
   Recipient: alert-recipient-1 (value=value)
""",
        )
        self.assert_pcs_fail(
            "alert recipient update alert-recipient value=value".split(),
            (
                "Error: Recipient 'value' in alert 'alert' already exists, "
                "use --force to override\n" + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success(
            "alert recipient update alert-recipient value=value --force".split(),
            "Warning: Recipient 'value' in alert 'alert' already exists\n",
        )
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=value)
   Recipient: alert-recipient-1 (value=value)
""",
        )

    def test_value_same_as_previous(self):
        self.assert_pcs_success("alert create path=test".split())
        self.assert_pcs_success(
            "alert recipient add alert value=rec_value".split()
        )
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=rec_value)
""",
        )
        self.assert_pcs_success(
            "alert recipient update alert-recipient value=rec_value".split()
        )
        self.assert_pcs_success(
            "alert config".split(),
            """\
Alerts:
 Alert: alert (path=test)
  Recipients:
   Recipient: alert-recipient (value=rec_value)
""",
        )

    def test_no_recipient(self):
        self.assert_pcs_fail(
            "alert recipient update rec description=desc".split(),
            "Error: recipient 'rec' does not exist\n",
        )

    def test_empty_value(self):
        self.assert_pcs_success("alert create path=test".split())
        self.assert_pcs_success(
            "alert recipient add alert value=rec_value id=rec".split()
        )
        self.assert_pcs_fail(
            "alert recipient update rec value=".split(),
            "Error: Recipient value '' is not valid.\n",
        )


class DeleteRemoveRecipientTest(PcsAlertTest):
    command = None

    def _test_usage(self):
        self.assert_pcs_fail(
            ["alert", "recipient", self.command],
            stdout_start=outdent(
                f"""
                Usage: pcs alert <command>
                    recipient {self.command} <"""
            ),
        )

    def _test_one(self):
        self.assert_pcs_success("alert create path=test".split())
        self.assert_pcs_success(
            "alert recipient add alert value=rec_value id=rec".split()
        )
        self.assert_pcs_success(
            "alert config".split(),
            outdent(
                """\
                Alerts:
                 Alert: alert (path=test)
                  Recipients:
                   Recipient: rec (value=rec_value)
                """
            ),
        )
        self.assert_pcs_success(["alert", "recipient", self.command, "rec"])
        self.assert_pcs_success(
            "alert config".split(),
            outdent(
                """\
                Alerts:
                 Alert: alert (path=test)
                """
            ),
        )

    def _test_multiple(self):
        self.assert_pcs_success("alert create path=test id=alert1".split())
        self.assert_pcs_success("alert create path=test id=alert2".split())
        self.assert_pcs_success(
            "alert recipient add alert1 value=rec_value1 id=rec1".split()
        )
        self.assert_pcs_success(
            "alert recipient add alert1 value=rec_value2 id=rec2".split()
        )
        self.assert_pcs_success(
            "alert recipient add alert2 value=rec_value3 id=rec3".split()
        )
        self.assert_pcs_success(
            "alert recipient add alert2 value=rec_value4 id=rec4".split()
        )
        self.assert_pcs_success(
            "alert config".split(),
            outdent(
                """\
                Alerts:
                 Alert: alert1 (path=test)
                  Recipients:
                   Recipient: rec1 (value=rec_value1)
                   Recipient: rec2 (value=rec_value2)
                 Alert: alert2 (path=test)
                  Recipients:
                   Recipient: rec3 (value=rec_value3)
                   Recipient: rec4 (value=rec_value4)
                """
            ),
        )
        self.assert_pcs_success(
            ["alert", "recipient", self.command, "rec1", "rec2", "rec4"]
        )
        self.assert_pcs_success(
            "alert config".split(),
            outdent(
                """\
                Alerts:
                 Alert: alert1 (path=test)
                 Alert: alert2 (path=test)
                  Recipients:
                   Recipient: rec3 (value=rec_value3)
                """
            ),
        )

    def _test_no_recipient(self):
        self.assert_pcs_success("alert create path=test id=alert1".split())
        self.assert_pcs_success(
            "alert recipient add alert1 value=rec_value1 id=rec1".split()
        )
        self.assert_pcs_fail(
            ["alert", "recipient", self.command, "rec1", "rec2", "rec3"],
            (
                "Error: recipient 'rec2' does not exist\n"
                "Error: recipient 'rec3' does not exist\n"
                + ERRORS_HAVE_OCCURRED
            ),
        )
        self.assert_pcs_success(
            "alert config".split(),
            outdent(
                """\
                Alerts:
                 Alert: alert1 (path=test)
                  Recipients:
                   Recipient: rec1 (value=rec_value1)
                """
            ),
        )


class DeleteRecipientTest(
    DeleteRemoveRecipientTest, metaclass=ParametrizedTestMetaClass
):
    command = "delete"


class RemoveRecipientTest(
    DeleteRemoveRecipientTest, metaclass=ParametrizedTestMetaClass
):
    command = "remove"



import grp
import pwd
from collections import namedtuple
from concurrent.futures import ProcessPoolExecutor
from ctypes import (
    CDLL,
    CFUNCTYPE,
    POINTER,
    Structure,
    byref,
    c_char,
    c_char_p,
    c_int,
    c_uint,
    c_void_p,
    cast,
    sizeof,
)
from ctypes.util import find_library

from tornado.gen import coroutine

from pcs.daemon import log

# pylint: disable=invalid-name, too-few-public-methods


class pam_message(Structure):
    _fields_ = [("msg_style", c_int), ("msg", POINTER(c_char))]


class pam_response(Structure):
    _fields_ = [("resp", POINTER(c_char)), ("resp_retcode", c_int)]


class pam_handle(Structure):
    _fields_ = [("handle", c_void_p)]


pam_conversation = CFUNCTYPE(
    c_int,  # return value type
    c_int,  # num_msg
    POINTER(POINTER(pam_message)),  # msg
    POINTER(POINTER(pam_response)),  # resp
    c_void_p,  # appdata_ptr
)


class pam_conv(Structure):
    _fields_ = [("conv", pam_conversation), ("appdata_ptr", c_void_p)]


def prep_fn(fn, restype, argtypes):
    fn.restype = restype
    fn.argtypes = argtypes
    return fn


# c_char_p represents the C char * datatype when it points to a zero-terminated
# string. For a general character pointer that may also point to binary data,
# POINTER(c_char) must be used... and it is used for strdup
libc = CDLL(find_library("c"))
libpam = CDLL(find_library("pam"))
strdup = prep_fn(libc.strdup, POINTER(c_char), [c_char_p])
calloc = prep_fn(libc.calloc, c_void_p, [c_uint, c_uint])
pam_authenticate = prep_fn(libpam.pam_authenticate, c_int, [pam_handle, c_int])
pam_acct_mgmt = prep_fn(libpam.pam_acct_mgmt, c_int, [pam_handle, c_int])
pam_end = prep_fn(libpam.pam_end, c_int, [pam_handle, c_int])
pam_start = prep_fn(
    libpam.pam_start,
    c_int,
    [c_char_p, c_char_p, POINTER(pam_conv), POINTER(pam_handle)],
)

PAM_SUCCESS = 0
PAM_PROMPT_ECHO_OFF = 1
PCSD_SERVICE = "pcsd"
HA_ADM_GROUP = "haclient"  # TODO use pacemaker_gname (??? + explanation why)


def authenticate_by_pam(username, password):
    @pam_conversation
    def conv(
        num_msg, msg, resp, appdata_ptr
    ):  # pylint: disable=unused-argument
        # it is: *resp = (pam_response *) calloc(num_msg, sizeof(pam_response))
        resp[0] = cast(
            calloc(num_msg, sizeof(pam_response)), POINTER(pam_response)
        )
        for i in range(num_msg):
            if msg[i].contents.msg_style == PAM_PROMPT_ECHO_OFF:
                resp.contents[i].resp = strdup(password.encode("utf8"))
                resp.contents[i].resp_retcode = 0
        return 0

    pamh = pam_handle()
    conversation = pam_conv(conv)
    returncode = pam_start(
        PCSD_SERVICE.encode("utf8"),
        username.encode("utf8"),
        byref(conversation),
        byref(pamh),
    )
    if returncode == PAM_SUCCESS:
        returncode = pam_authenticate(pamh, 0)
    if returncode == PAM_SUCCESS:
        returncode = pam_acct_mgmt(pamh, 0)
    pam_end(pamh, returncode)
    return returncode == PAM_SUCCESS


def get_user_groups_sync(username):
    return tuple(
        [group.gr_name for group in grp.getgrall() if username in group.gr_mem]
        + [
            # `group.gr_mem` does not contain the `username` when `group` is
            # primary for the `username` (the same is in /etc/group). So it is
            # necessary to add the primary group.
            grp.getgrgid(pwd.getpwnam(username).pw_gid).gr_name
        ]
    )


UserAuthInfo = namedtuple("UserAuthInfo", "name groups is_authorized")


class LoginLogger:
    @staticmethod
    def unable_determine_groups(username, e):
        log.pcsd.info(
            "Unable to determine groups of user '%s': %s", username, e
        )
        log.pcsd.info(
            "Failed login by '%s' (unable to determine user's groups)", username
        )

    @staticmethod
    def not_ha_adm_member(username, ha_adm_group):
        log.pcsd.info(
            "Failed login by '%s' (user is not a member of '%s' group)",
            username,
            ha_adm_group,
        )

    @staticmethod
    def success(username):
        log.pcsd.info("Successful login by '%s'", username)


class PlainLogger:
    @staticmethod
    def unable_determine_groups(username, e):
        log.pcsd.info(
            "Unable to determine groups of user '%s': %s", username, e
        )

    @staticmethod
    def not_ha_adm_member(username, ha_adm_group):
        log.pcsd.info(
            "User '%s' is not a member of '%s' group", username, ha_adm_group
        )

    def success(self, username):
        pass


def check_user_groups_sync(username, logger) -> UserAuthInfo:
    try:
        groups = get_user_groups_sync(username)
    except KeyError as e:
        logger.unable_determine_groups(username, e)
        return UserAuthInfo(username, [], is_authorized=False)

    if HA_ADM_GROUP not in groups:
        logger.not_ha_adm_member(username, HA_ADM_GROUP)
        return UserAuthInfo(username, groups, is_authorized=False)

    logger.success(username)
    return UserAuthInfo(username, groups, is_authorized=True)


def authorize_user_sync(username, password) -> UserAuthInfo:
    log.pcsd.info("Attempting login by '%s'", username)

    if not authenticate_by_pam(username, password):
        log.pcsd.info(
            "Failed login by '%s' (bad username or password)", username
        )
        return UserAuthInfo(username, [], is_authorized=False)

    return check_user_groups_sync(username, LoginLogger())


# TODO async/await version - how to do it?
# When async/await is used then the problem is:
# "TypeError: object Future can't be used in 'await' expression" is raised even
# if the function "convert_yielded" is used according to
# http://www.tornadoweb.org/en/stable/guide/coroutines.html#python-3-5-async-and-await
@coroutine
def run_in_process(sync_fn, *args):
    with ProcessPoolExecutor(max_workers=1) as pool:
        result = yield pool.submit(sync_fn, *args)
    return result


@coroutine
def authorize_user(username, password) -> UserAuthInfo:
    user = yield run_in_process(authorize_user_sync, username, password)
    return user


@coroutine
def check_user_groups(username) -> UserAuthInfo:
    user = yield run_in_process(check_user_groups_sync, username, PlainLogger())
    return user



from typing import (
    Any,
    Container,
    Iterable,
    List,
    Mapping,
    Tuple,
    cast,
)

from pcs.common import (
    file_type_codes,
    reports,
)
from pcs.common.dr import (
    DrConfigDto,
    DrConfigNodeDto,
    DrConfigSiteDto,
    DrSiteStatusDto,
)
from pcs.common.file import RawFileError
from pcs.common.interface import dto
from pcs.common.node_communicator import RequestTarget
from pcs.common.reports import ReportItemList
from pcs.common.reports import codes as report_codes
from pcs.common.reports.item import ReportItem
from pcs.lib import node_communication_format
from pcs.lib.communication.corosync import GetCorosyncConf
from pcs.lib.communication.nodes import (
    DistributeFilesWithoutForces,
    RemoveFilesWithoutForces,
)
from pcs.lib.communication.status import GetFullClusterStatusPlaintext
from pcs.lib.communication.tools import run as run_com_cmd
from pcs.lib.communication.tools import run_and_raise
from pcs.lib.corosync.config_facade import ConfigFacade as CorosyncConfigFacade
from pcs.lib.dr.config.facade import DrRole
from pcs.lib.dr.config.facade import Facade as DrConfigFacade
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.file.instance import FileInstance
from pcs.lib.file.raw_file import raw_file_error_report
from pcs.lib.file.toolbox import for_file_type as get_file_toolbox
from pcs.lib.interface.config import ParserErrorException
from pcs.lib.node import get_existing_nodes_names


def get_config(env: LibraryEnvironment) -> Mapping[str, Any]:
    """
    Return local disaster recovery config

    env -- LibraryEnvironment
    """
    report_processor = env.report_processor
    report_list, dr_config = _load_dr_config(env.get_dr_env().config)
    report_processor.report_list(report_list)
    if report_processor.has_errors:
        raise LibraryError()

    return dto.to_dict(
        DrConfigDto(
            local_site=DrConfigSiteDto(
                site_role=dr_config.local_role,
                node_list=[],
            ),
            remote_site_list=[
                DrConfigSiteDto(
                    site_role=site.role,
                    node_list=[
                        DrConfigNodeDto(name=name)
                        for name in site.node_name_list
                    ],
                )
                for site in dr_config.get_remote_site_list()
            ],
        )
    )


def set_recovery_site(env: LibraryEnvironment, node_name: str) -> None:
    """
    Set up disaster recovery with the local cluster being the primary site

    env
    node_name -- a known host from the recovery site
    """
    # pylint: disable=too-many-locals
    if env.ghost_file_codes:
        raise LibraryError(
            ReportItem.error(
                reports.messages.LiveEnvironmentRequired(env.ghost_file_codes)
            )
        )
    report_processor = env.report_processor
    dr_env = env.get_dr_env()
    if dr_env.config.raw_file.exists():
        report_processor.report(
            ReportItem.error(reports.messages.DrConfigAlreadyExist())
        )
    target_factory = env.get_node_target_factory()

    local_nodes, report_list = get_existing_nodes_names(
        env.get_corosync_conf(), error_on_missing_name=True
    )
    report_processor.report_list(report_list)

    if node_name in local_nodes:
        report_processor.report(
            ReportItem.error(reports.messages.NodeInLocalCluster(node_name))
        )

    report_list, local_targets = target_factory.get_target_list_with_reports(
        local_nodes, allow_skip=False, report_none_host_found=False
    )
    report_processor.report_list(report_list)

    report_list, remote_targets = target_factory.get_target_list_with_reports(
        [node_name], allow_skip=False, report_none_host_found=False
    )
    report_processor.report_list(report_list)

    if report_processor.has_errors:
        raise LibraryError()

    # TODO The new file framework doesn't support network communication yet.
    com_cmd = GetCorosyncConf(env.report_processor)
    com_cmd.set_targets(remote_targets)
    corosync_conf_instance = FileInstance.for_corosync_conf()
    try:
        remote_cluster_nodes, report_list = get_existing_nodes_names(
            cast(
                CorosyncConfigFacade,
                corosync_conf_instance.raw_to_facade(
                    run_and_raise(env.get_node_communicator(), com_cmd).encode(
                        "utf-8"
                    )
                ),
            ),
            error_on_missing_name=True,
        )
    except ParserErrorException as e:
        report_processor.report_list(
            corosync_conf_instance.toolbox.parser.exception_to_report_list(
                e,
                file_type_codes.COROSYNC_CONF,
                None,
                force_code=None,
                is_forced_or_warning=False,
            )
        )

    if report_processor.report_list(report_list).has_errors:
        raise LibraryError()

    # ensure we have tokens for all nodes of remote cluster
    report_list, remote_targets = target_factory.get_target_list_with_reports(
        remote_cluster_nodes, allow_skip=False, report_none_host_found=False
    )
    if report_processor.report_list(report_list).has_errors:
        raise LibraryError()
    dr_config_exporter = get_file_toolbox(
        file_type_codes.PCS_DR_CONFIG
    ).exporter
    # create dr config for remote cluster
    remote_dr_cfg = dr_env.create_facade(DrRole.RECOVERY)
    remote_dr_cfg.add_site(DrRole.PRIMARY, local_nodes)
    # send config to all node of remote cluster
    distribute_file_cmd = DistributeFilesWithoutForces(
        env.report_processor,
        node_communication_format.pcs_dr_config_file(
            dr_config_exporter.export(remote_dr_cfg.config)
        ),
    )
    distribute_file_cmd.set_targets(remote_targets)
    run_and_raise(env.get_node_communicator(), distribute_file_cmd)
    # create new dr config, with local cluster as primary site
    local_dr_cfg = dr_env.create_facade(DrRole.PRIMARY)
    local_dr_cfg.add_site(DrRole.RECOVERY, remote_cluster_nodes)
    distribute_file_cmd = DistributeFilesWithoutForces(
        env.report_processor,
        node_communication_format.pcs_dr_config_file(
            dr_config_exporter.export(local_dr_cfg.config)
        ),
    )
    distribute_file_cmd.set_targets(local_targets)
    run_and_raise(env.get_node_communicator(), distribute_file_cmd)
    # Note: No token sync across multiple clusters. Most probably they are in
    # different subnetworks.


def status_all_sites_plaintext(
    env: LibraryEnvironment,
    hide_inactive_resources: bool = False,
    verbose: bool = False,
) -> List[Mapping[str, Any]]:
    """
    Return local site's and all remote sites' status as plaintext

    env -- LibraryEnvironment
    hide_inactive_resources -- if True, do not display non-running resources
    verbose -- if True, display more info
    """
    # The command does not provide an option to skip offline / unreachable /
    # misbehaving nodes.
    # The point of such skipping is to stop a command if it is unable to make
    # changes on all nodes. The user can then decide to proceed anyway and
    # make changes on the skipped nodes later manually.
    # This command only reads from nodes so it automatically asks other nodes
    # if one is offline / misbehaving.
    class SiteData:
        def __init__(
            self,
            local: bool,
            role: DrRole,
            target_list: Iterable[RequestTarget],
        ) -> None:
            self.local = local
            self.role = role
            self.target_list = target_list
            self.status_loaded = False
            self.status_plaintext = ""

    if env.ghost_file_codes:
        raise LibraryError(
            ReportItem.error(
                reports.messages.LiveEnvironmentRequired(env.ghost_file_codes)
            )
        )

    report_processor = env.report_processor
    report_list, dr_config = _load_dr_config(env.get_dr_env().config)
    report_processor.report_list(report_list)
    if report_processor.has_errors:
        raise LibraryError()

    site_data_list = []
    target_factory = env.get_node_target_factory()

    # get local nodes
    local_nodes, report_list = get_existing_nodes_names(env.get_corosync_conf())
    report_processor.report_list(report_list)
    report_list, local_targets = target_factory.get_target_list_with_reports(
        local_nodes,
        skip_non_existing=True,
    )
    report_processor.report_list(report_list)
    site_data_list.append(SiteData(True, dr_config.local_role, local_targets))

    # get remote sites' nodes
    for conf_remote_site in dr_config.get_remote_site_list():
        (
            report_list,
            remote_targets,
        ) = target_factory.get_target_list_with_reports(
            conf_remote_site.node_name_list,
            skip_non_existing=True,
        )
        report_processor.report_list(report_list)
        site_data_list.append(
            SiteData(False, conf_remote_site.role, remote_targets)
        )
    if report_processor.has_errors:
        raise LibraryError()

    # get all statuses
    for site_data in site_data_list:
        com_cmd = GetFullClusterStatusPlaintext(
            report_processor,
            hide_inactive_resources=hide_inactive_resources,
            verbose=verbose,
        )
        com_cmd.set_targets(site_data.target_list)
        site_data.status_loaded, site_data.status_plaintext = run_com_cmd(
            env.get_node_communicator(), com_cmd
        )

    return [
        dto.to_dict(
            DrSiteStatusDto(
                local_site=site_data.local,
                site_role=site_data.role,
                status_plaintext=site_data.status_plaintext,
                status_successfully_obtained=site_data.status_loaded,
            )
        )
        for site_data in site_data_list
    ]


def _load_dr_config(
    config_file: FileInstance,
) -> Tuple[ReportItemList, DrConfigFacade]:
    if not config_file.raw_file.exists():
        return (
            [
                ReportItem.error(reports.messages.DrConfigDoesNotExist()),
            ],
            DrConfigFacade.empty(),
        )
    try:
        return [], cast(DrConfigFacade, config_file.read_to_facade())
    except RawFileError as e:
        return [raw_file_error_report(e)], DrConfigFacade.empty()
    except ParserErrorException as e:
        return (
            config_file.parser_exception_to_report_list(e),
            DrConfigFacade.empty(),
        )


def destroy(
    env: LibraryEnvironment,
    force_flags: Container[reports.types.ForceCode] = (),
) -> None:
    """
    Destroy disaster-recovery configuration on all sites
    """
    if env.ghost_file_codes:
        raise LibraryError(
            ReportItem.error(
                reports.messages.LiveEnvironmentRequired(env.ghost_file_codes)
            )
        )

    report_processor = env.report_processor
    skip_offline = report_codes.SKIP_OFFLINE_NODES in force_flags

    report_list, dr_config = _load_dr_config(env.get_dr_env().config)
    report_processor.report_list(report_list)

    if report_processor.has_errors:
        raise LibraryError()

    local_nodes, report_list = get_existing_nodes_names(env.get_corosync_conf())
    report_processor.report_list(report_list)

    if report_processor.has_errors:
        raise LibraryError()

    remote_nodes: List[str] = []
    for conf_remote_site in dr_config.get_remote_site_list():
        remote_nodes.extend(conf_remote_site.node_name_list)

    target_factory = env.get_node_target_factory()
    report_list, targets = target_factory.get_target_list_with_reports(
        remote_nodes + local_nodes,
        skip_non_existing=skip_offline,
    )
    report_processor.report_list(report_list)
    if report_processor.has_errors:
        raise LibraryError()

    com_cmd = RemoveFilesWithoutForces(
        env.report_processor,
        {
            "pcs disaster-recovery config": {
                "type": "pcs_disaster_recovery_conf",
            },
        },
    )
    com_cmd.set_targets(targets)
    run_and_raise(env.get_node_communicator(), com_cmd)



import fcntl
import os
import shutil
from typing import (
    NewType,
    Optional,
)

from dataclasses import dataclass

from pcs.common.file_type_codes import FileTypeCode
from pcs.common.tools import format_os_error

# TODO add logging (logger / debug reports ?) to the RawFile class; be aware
# the class is used both in pcs.cli and pcs.lib packages


@dataclass(frozen=True)
class FileMetadata:
    file_type_code: FileTypeCode
    path: str
    owner_user_name: Optional[str]
    owner_group_name: Optional[str]
    permissions: Optional[int]
    is_binary: bool


FileAction = NewType("FileAction", str)


class RawFileError(Exception):
    # So far there has been no need to have a separate exception for each
    # action. Actions must be passed in a report and we certainely do not want
    # a separate report for each action.

    ACTION_CHMOD = FileAction("chmod")
    ACTION_CHOWN = FileAction("chown")
    ACTION_READ = FileAction("read")
    ACTION_REMOVE = FileAction("remove")
    ACTION_WRITE = FileAction("write")

    def __init__(
        self, metadata: FileMetadata, action: FileAction, reason: str = ""
    ):
        """
        metadata -- describes the file involved in the error
        action -- possible values enumerated in RawFileError
        reason -- plain text error details
        """
        super().__init__()
        self.metadata = metadata
        self.action = action
        self.reason = reason


class FileAlreadyExists(RawFileError):
    def __init__(self, metadata: FileMetadata):
        """
        metadata -- describes the file involved in the error
        """
        super().__init__(metadata, RawFileError.ACTION_WRITE)


class RawFileInterface:
    def __init__(self, metadata: FileMetadata):
        """
        metadata -- describes the file and provides its metadata
        """
        self.__metadata = metadata

    @property
    def metadata(self) -> FileMetadata:
        return self.__metadata

    def exists(self) -> bool:
        """
        Return True if file exists, False otherwise
        """
        raise NotImplementedError()

    def read(self) -> bytes:
        """
        Return content of the file as bytes
        """
        raise NotImplementedError()

    def write(self, file_data: bytes, can_overwrite: bool = False) -> None:
        """
        Write file_data to the file

        file_data -- data to be written
        can_overwrite -- raise if False and the file already exists
        """
        raise NotImplementedError()


class RawFile(RawFileInterface):
    def exists(self) -> bool:
        # Returns False if the file is not accessible, does not raise.
        return os.path.exists(self.metadata.path)

    def read(self) -> bytes:
        try:
            mode = "rb" if self.metadata.is_binary else "r"
            with open(self.metadata.path, mode) as my_file:
                # the lock is released when the file gets closed on leaving the
                # with statement
                fcntl.flock(my_file.fileno(), fcntl.LOCK_SH)
                content = my_file.read()
                return (
                    content
                    if self.metadata.is_binary
                    else content.encode("utf-8")
                )
        except OSError as e:
            # Specific exception if the file does not exist is not needed,
            # anyone can and should check that using the exists method.
            raise RawFileError(
                self.metadata, RawFileError.ACTION_READ, format_os_error(e)
            ) from e

    def write(self, file_data: bytes, can_overwrite: bool = False) -> None:
        try:
            mode = "{write_mode}{binary_mode}".format(
                write_mode="w" if can_overwrite else "x",
                binary_mode="b" if self.metadata.is_binary else "",
            )
            # It seems pylint cannot process constructing the mode variable and
            # gives a false positive.
            # pylint: disable=bad-open-mode
            with open(self.metadata.path, mode) as my_file:
                # the lock is released when the file gets closed on leaving the
                # with statement
                fcntl.flock(my_file.fileno(), fcntl.LOCK_EX)
                # Set the ownership and permissions to cover the case when we
                # just created the file. If the file already existed, make sure
                # the ownership and permissions are correct before writing any
                # data into it.
                if (
                    self.metadata.owner_user_name is not None
                    or self.metadata.owner_group_name is not None
                ):
                    try:
                        shutil.chown(
                            self.metadata.path,
                            self.metadata.owner_user_name,
                            self.metadata.owner_group_name,
                        )
                    except LookupError as e:
                        raise RawFileError(
                            self.metadata, RawFileError.ACTION_CHOWN, str(e)
                        ) from e
                    except OSError as e:
                        raise RawFileError(
                            self.metadata,
                            RawFileError.ACTION_CHOWN,
                            format_os_error(e),
                        ) from e

                if self.metadata.permissions is not None:
                    try:
                        os.chmod(my_file.fileno(), self.metadata.permissions)
                    except OSError as e:
                        raise RawFileError(
                            self.metadata,
                            RawFileError.ACTION_CHMOD,
                            format_os_error(e),
                        ) from e
                # Write file data
                my_file.write(
                    file_data
                    if self.metadata.is_binary
                    else file_data.decode("utf-8")
                )
        except FileExistsError as e:
            raise FileAlreadyExists(self.metadata) from e
        except OSError as e:
            raise RawFileError(
                self.metadata, RawFileError.ACTION_WRITE, format_os_error(e)
            ) from e

    def remove(self, fail_if_file_not_found: bool = True) -> None:
        try:
            os.remove(self.metadata.path)
        except FileNotFoundError as e:
            if fail_if_file_not_found:
                raise self.__get_raw_file_error(e) from e
        except OSError as e:
            raise self.__get_raw_file_error(e) from e

    def backup(self) -> None:
        # TODO implement
        raise NotImplementedError()

    def __get_raw_file_error(self, e: OSError) -> RawFileError:
        return RawFileError(
            self.metadata, RawFileError.ACTION_REMOVE, format_os_error(e)
        )



from tornado.web import (
    Finish,
    StaticFileHandler,
)

from pcs.daemon.app.common import EnhanceHeadersMixin


class AjaxMixin:
    """
    AjaxMixin adds methods for an ajax request detection and common unauthorized
    response.
    """

    @property
    def is_ajax(self):
        return (
            self.request.headers.get("X-Requested-With", default=None)
            == "XMLHttpRequest"
        )

    def unauthorized(self):
        self.set_status(401)
        self.write('{"notauthorized":"true"}')
        return Finish()


class StaticFile(EnhanceHeadersMixin, StaticFileHandler):
    # abstract method `data_received` does need to be overridden. This
    # method should be implemented to handle streamed request data.
    # BUT static files are not streamed SO:
    # pylint: disable=abstract-method
    def initialize(self, path, default_filename=None):
        super().initialize(path, default_filename)
        # allow static files to be cached
        self.clear_header_cache_control()



import os.path
from unittest import TestCase

from pcs import settings
from pcs.common import reports
from pcs.common.pacemaker.defaults import CibDefaultsDto
from pcs.common.pacemaker.nvset import (
    CibNvpairDto,
    CibNvsetDto,
)
from pcs.common.pacemaker.rule import (
    CibRuleDateCommonDto,
    CibRuleExpressionDto,
)
from pcs.common.types import (
    CibRuleExpressionType,
    CibRuleInEffectStatus,
)
from pcs.lib.commands import cib_options

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.command_env.config_runner_pcmk import (
    RULE_EXPIRED_RETURNCODE,
    RULE_IN_EFFECT_RETURNCODE,
    RULE_NOT_YET_IN_EFFECT_RETURNCODE,
)


class DefaultsCreateMixin:
    command = lambda *args, **kwargs: None
    tag = ""

    def setUp(self):
        # pylint: disable=invalid-name
        self.env_assist, self.config = get_env_tools(self)
        self.config.runner.cib.load()

    def test_success_minimal(self):
        defaults_xml = f"""
            <{self.tag}>
                <meta_attributes id="{self.tag}-meta_attributes" />
            </{self.tag}>
        """
        self.config.env.push_cib(optional_in_conf=defaults_xml)

        self.command(self.env_assist.get_env(), {}, {})

        self.env_assist.assert_reports(
            [fixture.warn(reports.codes.DEFAULTS_CAN_BE_OVERRIDDEN)]
        )

    def test_success_one_set_already_there(self):
        defaults_xml_1 = f"""
            <{self.tag}>
                <meta_attributes id="{self.tag}-meta_attributes" />
            </{self.tag}>
        """
        defaults_xml_2 = f"""
            <{self.tag}>
                <meta_attributes id="{self.tag}-meta_attributes" />
                <meta_attributes id="{self.tag}-meta_attributes-1" />
            </{self.tag}>
        """
        self.config.runner.cib.load(
            instead="runner.cib.load", optional_in_conf=defaults_xml_1
        )
        self.config.env.push_cib(optional_in_conf=defaults_xml_2)

        self.command(self.env_assist.get_env(), {}, {})

        self.env_assist.assert_reports(
            [fixture.warn(reports.codes.DEFAULTS_CAN_BE_OVERRIDDEN)]
        )

    def test_success_cib_upgrade_rsc_rules(self):
        defaults_xml = f"""
            <{self.tag}>
                <meta_attributes id="{self.tag}-meta_attributes">
                    <rule id="{self.tag}-meta_attributes-rule"
                        boolean-op="and" score="INFINITY"
                    >
                        <rsc_expression
                            id="{self.tag}-meta_attributes-rule-rsc-ocf-pacemaker-Dummy"
                            class="ocf" provider="pacemaker" type="Dummy"
                        />
                    </rule>
                </meta_attributes>
            </{self.tag}>
        """
        self.config.runner.cib.load(
            name="load_cib_old_version",
            filename="cib-empty-3.3.xml",
            before="runner.cib.load",
        )
        self.config.runner.cib.upgrade(before="runner.cib.load")
        self.config.runner.cib.load(
            filename="cib-empty-3.4.xml", instead="runner.cib.load"
        )
        self.config.env.push_cib(optional_in_conf=defaults_xml)

        self.command(
            self.env_assist.get_env(),
            {},
            {},
            nvset_rule="resource ocf:pacemaker:Dummy",
        )

        self.env_assist.assert_reports(
            [
                fixture.info(reports.codes.CIB_UPGRADE_SUCCESSFUL),
                fixture.warn(reports.codes.DEFAULTS_CAN_BE_OVERRIDDEN),
            ]
        )

    def test_validation(self):
        self.config.runner.cib.load(
            filename="cib-empty-3.4.xml", instead="runner.cib.load"
        )
        self.env_assist.assert_raise_library_error(
            lambda: self.command(
                self.env_assist.get_env(),
                {},
                {"unknown-option": "value"},
                "bad rule",
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    force_code=reports.codes.FORCE,
                    option_names=["unknown-option"],
                    allowed=["id", "score"],
                    option_type=None,
                    allowed_patterns=[],
                ),
                fixture.error(
                    reports.codes.RULE_EXPRESSION_PARSE_ERROR,
                    rule_string="bad rule",
                    reason="Expected 'eq'",
                    rule_line="bad rule",
                    line_number=1,
                    column_number=5,
                    position=4,
                ),
            ]
        )

    def test_validation_forced(self):
        defaults_xml = f"""
            <{self.tag}>
                <meta_attributes id="{self.tag}-meta_attributes"
                    unknown-option="value"
                />
            </{self.tag}>
        """
        self.config.env.push_cib(optional_in_conf=defaults_xml)

        self.command(
            self.env_assist.get_env(),
            {},
            {"unknown-option": "value"},
            force_flags={reports.codes.FORCE},
        )

        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["unknown-option"],
                    allowed=["id", "score"],
                    option_type=None,
                    allowed_patterns=[],
                ),
                fixture.warn(reports.codes.DEFAULTS_CAN_BE_OVERRIDDEN),
            ]
        )


class ResourceDefaultsCreate(DefaultsCreateMixin, TestCase):
    command = staticmethod(cib_options.resource_defaults_create)
    tag = "rsc_defaults"

    def test_rule_op_expression_not_allowed(self):
        self.config.runner.cib.load(
            filename="cib-empty-3.4.xml", instead="runner.cib.load"
        )
        self.env_assist.assert_raise_library_error(
            lambda: self.command(
                self.env_assist.get_env(),
                {},
                {},
                "op monitor",
                force_flags={reports.codes.FORCE},
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.RULE_EXPRESSION_NOT_ALLOWED,
                    expression_type=CibRuleExpressionType.OP_EXPRESSION,
                ),
            ]
        )

    def test_rule_node_attr_expression_not_allowed(self):
        self.config.runner.cib.load(
            filename="cib-empty-3.4.xml", instead="runner.cib.load"
        )
        self.env_assist.assert_raise_library_error(
            lambda: self.command(
                self.env_assist.get_env(),
                {},
                {},
                "defined attr",
                force_flags={reports.codes.FORCE},
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.RULE_EXPRESSION_NOT_ALLOWED,
                    expression_type=CibRuleExpressionType.EXPRESSION,
                ),
            ]
        )

    def test_success_full(self):
        defaults_xml = f"""
            <{self.tag}>
                <meta_attributes id="my-id" score="10">
                    <rule id="my-id-rule" boolean-op="and" score="INFINITY">
                        <rsc_expression id="my-id-rule-rsc-ocf-pacemaker-Dummy"
                            class="ocf" provider="pacemaker" type="Dummy"
                        />
                        <rule id="my-id-rule-rule" boolean-op="or" score="0">
                            <date_expression id="my-id-rule-rule-expr"
                                operation="lt" end="2020-08-07"
                            />
                            <date_expression id="my-id-rule-rule-expr-1"
                                operation="in_range"
                                start="2020-09-01" end="2020-09-11"
                            />
                            <date_expression id="my-id-rule-rule-expr-2"
                                operation="in_range" start="2020-10-01"
                            >
                                <duration id="my-id-rule-rule-expr-2-duration"
                                    months="1"
                                />
                            </date_expression>
                            <date_expression id="my-id-rule-rule-expr-3"
                                operation="date_spec"
                            >
                                <date_spec id="my-id-rule-rule-expr-3-datespec"
                                    years="2021-2022"
                                />
                            </date_expression>
                            <date_expression id="my-id-rule-rule-expr-4"
                                operation="in_range" end="2020-12-11"
                            />
                        </rule>
                    </rule>
                    <nvpair id="my-id-name1" name="name1" value="value1" />
                    <nvpair id="my-id-2name" name="2na#me" value="value2" />
                </meta_attributes>
            </{self.tag}>
        """
        self.config.runner.cib.load(
            filename="cib-empty-3.4.xml", instead="runner.cib.load"
        )
        self.config.env.push_cib(optional_in_conf=defaults_xml)

        self.command(
            self.env_assist.get_env(),
            {"name1": "value1", "2na#me": "value2"},
            {"id": "my-id", "score": "10"},
            nvset_rule=(
                "resource ocf:pacemaker:Dummy and "
                "(date lt 2020-08-07 or "
                "date in_range 2020-09-01 to 2020-09-11 or "
                "date in_range 2020-10-01 to duration months=1 or "
                "date-spec years=2021-2022 or "
                "date in_range to 2020-12-11)"
            ),
        )

        self.env_assist.assert_reports(
            [fixture.warn(reports.codes.DEFAULTS_CAN_BE_OVERRIDDEN)]
        )


class OperationDefaultsCreate(DefaultsCreateMixin, TestCase):
    command = staticmethod(cib_options.operation_defaults_create)
    tag = "op_defaults"

    def test_success_full(self):
        defaults_xml = f"""
            <{self.tag}>
                <meta_attributes id="my-id" score="10">
                    <rule id="my-id-rule" boolean-op="and" score="INFINITY">
                        <rsc_expression id="my-id-rule-rsc-ocf-pacemaker-Dummy"
                            class="ocf" provider="pacemaker" type="Dummy"
                        />
                        <op_expression id="my-id-rule-op-monitor" name="monitor"
                            interval="30"
                        />
                        <rule id="my-id-rule-rule" boolean-op="or" score="0">
                            <expression id="my-id-rule-rule-expr"
                                operation="defined" attribute="attr1"
                            />
                            <expression id="my-id-rule-rule-expr-1"
                                attribute="attr2" operation="gt"
                                type="number" value="5"
                            />
                            <date_expression id="my-id-rule-rule-expr-2"
                                operation="lt" end="2020-08-07"
                            />
                            <date_expression id="my-id-rule-rule-expr-3"
                                operation="in_range"
                                start="2020-09-01" end="2020-09-11"
                            />
                            <date_expression id="my-id-rule-rule-expr-4"
                                operation="in_range" start="2020-10-01"
                            >
                                <duration id="my-id-rule-rule-expr-4-duration"
                                    months="1"
                                />
                            </date_expression>
                            <date_expression id="my-id-rule-rule-expr-5"
                                operation="date_spec"
                            >
                                <date_spec id="my-id-rule-rule-expr-5-datespec"
                                    years="2021-2022"
                                />
                            </date_expression>
                            <date_expression id="my-id-rule-rule-expr-6"
                                operation="in_range" end="2020-12-11"
                            />
                        </rule>
                    </rule>
                    <nvpair id="my-id-name1" name="name1" value="value1" />
                    <nvpair id="my-id-2name" name="2na#me" value="value2" />
                </meta_attributes>
            </{self.tag}>
        """
        self.config.runner.cib.load(
            filename="cib-empty-3.4.xml", instead="runner.cib.load"
        )
        self.config.env.push_cib(optional_in_conf=defaults_xml)

        self.command(
            self.env_assist.get_env(),
            {"name1": "value1", "2na#me": "value2"},
            {"id": "my-id", "score": "10"},
            nvset_rule=(
                "resource ocf:pacemaker:Dummy and op monitor interval=30 and "
                "(defined attr1 or attr2 gt number 5 or date lt 2020-08-07 or "
                "date in_range 2020-09-01 to 2020-09-11 or "
                "date in_range 2020-10-01 to duration months=1 or "
                "date-spec years=2021-2022 or "
                "date in_range to 2020-12-11)"
            ),
        )

        self.env_assist.assert_reports(
            [fixture.warn(reports.codes.DEFAULTS_CAN_BE_OVERRIDDEN)]
        )

    def test_success_cib_upgrade_node_attr_type_int(self):
        defaults_xml = f"""
            <{self.tag}>
                <meta_attributes id="{self.tag}-meta_attributes">
                    <rule id="{self.tag}-meta_attributes-rule"
                        boolean-op="and" score="INFINITY"
                    >
                        <expression id="{self.tag}-meta_attributes-rule-expr"
                            attribute="attr" operation="eq" type="integer" value="5"
                        />
                    </rule>
                </meta_attributes>
            </{self.tag}>
        """
        self.config.runner.cib.load(
            name="load_cib_old_version",
            filename="cib-empty-3.3.xml",
            before="runner.cib.load",
        )
        self.config.runner.cib.upgrade(before="runner.cib.load")
        self.config.runner.cib.load(
            filename="cib-empty-3.5.xml", instead="runner.cib.load"
        )
        self.config.env.push_cib(optional_in_conf=defaults_xml)

        self.command(
            self.env_assist.get_env(),
            {},
            {},
            nvset_rule="attr eq integer 5",
        )

        self.env_assist.assert_reports(
            [
                fixture.info(reports.codes.CIB_UPGRADE_SUCCESSFUL),
                fixture.warn(reports.codes.DEFAULTS_CAN_BE_OVERRIDDEN),
            ]
        )

    def test_success_cib_upgrade_node_attr_type_int_not_upgraded(self):
        defaults_xml = f"""
            <{self.tag}>
                <meta_attributes id="{self.tag}-meta_attributes">
                    <rule id="{self.tag}-meta_attributes-rule"
                        boolean-op="and" score="INFINITY"
                    >
                        <expression id="{self.tag}-meta_attributes-rule-expr"
                            attribute="attr" operation="eq" type="number" value="5"
                        />
                    </rule>
                </meta_attributes>
            </{self.tag}>
        """
        self.config.runner.cib.load(
            name="load_cib_old_version",
            filename="cib-empty-3.3.xml",
            before="runner.cib.load",
        )
        self.config.runner.cib.upgrade(before="runner.cib.load")
        self.config.runner.cib.load(
            filename="cib-empty-3.4.xml", instead="runner.cib.load"
        )
        self.config.env.push_cib(optional_in_conf=defaults_xml)

        self.command(
            self.env_assist.get_env(),
            {},
            {},
            nvset_rule="attr eq integer 5",
        )

        self.env_assist.assert_reports(
            [
                fixture.info(reports.codes.CIB_UPGRADE_SUCCESSFUL),
                fixture.warn(reports.codes.DEFAULTS_CAN_BE_OVERRIDDEN),
            ]
        )

    def test_success_cib_upgrade_mixed(self):
        defaults_xml = f"""
            <{self.tag}>
                <meta_attributes id="{self.tag}-meta_attributes">
                    <rule id="{self.tag}-meta_attributes-rule"
                        boolean-op="and" score="INFINITY"
                    >
                        <op_expression
                            id="{self.tag}-meta_attributes-rule-op-monitor"
                            name="monitor"
                        />
                        <expression id="{self.tag}-meta_attributes-rule-expr"
                            attribute="attr" operation="eq" type="integer"
                            value="5"
                        />
                    </rule>
                </meta_attributes>
            </{self.tag}>
        """
        self.config.runner.cib.load(
            name="load_cib_old_version",
            filename="cib-empty-3.3.xml",
            before="runner.cib.load",
        )
        self.config.runner.cib.upgrade(before="runner.cib.load")
        self.config.runner.cib.load(
            filename="cib-empty-3.5.xml", instead="runner.cib.load"
        )
        self.config.env.push_cib(optional_in_conf=defaults_xml)

        self.command(
            self.env_assist.get_env(),
            {},
            {},
            nvset_rule="op monitor and attr eq integer 5",
        )

        self.env_assist.assert_reports(
            [
                fixture.info(reports.codes.CIB_UPGRADE_SUCCESSFUL),
                fixture.warn(reports.codes.DEFAULTS_CAN_BE_OVERRIDDEN),
            ]
        )


class DefaultsConfigMixin:
    command = lambda *args, **kwargs: None
    tag = ""

    def setUp(self):
        # pylint: disable=invalid-name
        self.env_assist, self.config = get_env_tools(self)

    @staticmethod
    def fixture_expired_dto(expired):
        return CibNvsetDto(
            "my-id",
            {},
            CibRuleExpressionDto(
                "my-id-rule",
                CibRuleExpressionType.RULE,
                expired,
                {"boolean-op": "and"},
                None,
                None,
                [
                    CibRuleExpressionDto(
                        "my-id-rule-expr",
                        CibRuleExpressionType.EXPRESSION,
                        CibRuleInEffectStatus.UNKNOWN,
                        {
                            "operation": "defined",
                            "attribute": "attr1",
                        },
                        None,
                        None,
                        [],
                        "defined attr1",
                    ),
                ],
                "defined attr1",
            ),
            [CibNvpairDto("my-id-pair1", "name1", "value1")],
        )

    def test_empty(self):
        defaults_xml = f"""<{self.tag} />"""
        self.config.runner.cib.load(
            filename="cib-empty-3.4.xml", optional_in_conf=defaults_xml
        )
        self.config.fs.isfile(
            (os.path.join(settings.pacemaker_binaries, "crm_rule")),
            return_value=True,
        )
        self.assertEqual(
            CibDefaultsDto(instance_attributes=[], meta_attributes=[]),
            self.command(self.env_assist.get_env(), True),
        )

    def test_full(self):
        defaults_xml = f"""
            <{self.tag}>
                <meta_attributes id="{self.tag}-meta_attributes">
                    <rule id="{self.tag}-meta_attributes-rule"
                        boolean-op="and" score="INFINITY"
                    >
                        <rsc_expression
                            id="{self.tag}-meta_attributes-rule-rsc-Dummy"
                            class="ocf" provider="pacemaker" type="Dummy"
                        />
                        <op_expression
                            id="{self.tag}-meta_attributes-rule-op-monitor"
                            name="monitor" interval="30"
                        />
                        <rule id="{self.tag}-meta_attributes-rule-rule"
                            boolean-op="or"
                        >
                            <expression
                                id="{self.tag}-meta_attributes-rule-rule-expr"
                                operation="defined" attribute="attr1"
                            />
                            <expression
                                id="{self.tag}-meta_attributes-rule-rule-expr-1"
                                attribute="attr2" operation="gt"
                                type="integer" value="5"
                            />
                            <date_expression
                                id="{self.tag}-meta_attributes-rule-rule-expr-2"
                                operation="lt" end="2020-08-07"
                            />
                            <date_expression
                                id="{self.tag}-meta_attributes-rule-rule-expr-3"
                                operation="in_range"
                                start="2020-09-01" end="2020-09-11"
                            />
                            <date_expression
                                id="{self.tag}-meta_attributes-rule-rule-expr-4"
                                operation="in_range" start="2020-10-01"
                            >
                                <duration
                                    id="{self.tag}-meta_attributes-rule-rule-expr-4-duration"
                                    months="1"
                                />
                            </date_expression>
                            <date_expression
                                id="{self.tag}-meta_attributes-rule-rule-expr-5"
                                operation="date_spec"
                            >
                                <date_spec
                                    id="{self.tag}-meta_attributes-rule-rule-expr-5-datespec"
                                    years="2021-2022"
                                />
                            </date_expression>
                            <date_expression
                                id="{self.tag}-meta_attributes-rule-rule-expr-6"
                                operation="in_range" end="2020-12-11"
                            />
                        </rule>
                    </rule>
                    <nvpair id="my-id-pair1" name="name1" value="value1" />
                    <nvpair id="my-id-pair2" name="name2" value="value2" />
                </meta_attributes>
                <instance_attributes id="instance">
                    <nvpair id="instance-pair" name="inst" value="ance" />
                </instance_attributes>
                <meta_attributes id="meta-plain" score="123">
                    <nvpair id="my-id-pair3" name="name1" value="value1" />
                </meta_attributes>
            </{self.tag}>
        """
        self.config.runner.cib.load(
            filename="cib-empty-3.4.xml", optional_in_conf=defaults_xml
        )
        self.config.fs.isfile(
            (os.path.join(settings.pacemaker_binaries, "crm_rule")),
            return_value=False,
        )
        self.assertEqual(
            CibDefaultsDto(
                instance_attributes=[
                    CibNvsetDto(
                        "instance",
                        {},
                        None,
                        [CibNvpairDto("instance-pair", "inst", "ance")],
                    ),
                ],
                meta_attributes=[
                    CibNvsetDto(
                        f"{self.tag}-meta_attributes",
                        {},
                        CibRuleExpressionDto(
                            f"{self.tag}-meta_attributes-rule",
                            CibRuleExpressionType.RULE,
                            CibRuleInEffectStatus.UNKNOWN,
                            {"boolean-op": "and", "score": "INFINITY"},
                            None,
                            None,
                            [
                                CibRuleExpressionDto(
                                    f"{self.tag}-meta_attributes-rule-rsc-Dummy",
                                    CibRuleExpressionType.RSC_EXPRESSION,
                                    CibRuleInEffectStatus.UNKNOWN,
                                    {
                                        "class": "ocf",
                                        "provider": "pacemaker",
                                        "type": "Dummy",
                                    },
                                    None,
                                    None,
                                    [],
                                    "resource ocf:pacemaker:Dummy",
                                ),
                                CibRuleExpressionDto(
                                    f"{self.tag}-meta_attributes-rule-op-monitor",
                                    CibRuleExpressionType.OP_EXPRESSION,
                                    CibRuleInEffectStatus.UNKNOWN,
                                    {
                                        "name": "monitor",
                                        "interval": "30",
                                    },
                                    None,
                                    None,
                                    [],
                                    "op monitor interval=30",
                                ),
                                CibRuleExpressionDto(
                                    f"{self.tag}-meta_attributes-rule-rule",
                                    CibRuleExpressionType.RULE,
                                    CibRuleInEffectStatus.UNKNOWN,
                                    {"boolean-op": "or"},
                                    None,
                                    None,
                                    [
                                        CibRuleExpressionDto(
                                            f"{self.tag}-meta_attributes-rule-rule-expr",
                                            CibRuleExpressionType.EXPRESSION,
                                            CibRuleInEffectStatus.UNKNOWN,
                                            {
                                                "operation": "defined",
                                                "attribute": "attr1",
                                            },
                                            None,
                                            None,
                                            [],
                                            "defined attr1",
                                        ),
                                        CibRuleExpressionDto(
                                            f"{self.tag}-meta_attributes-rule-rule-expr-1",
                                            CibRuleExpressionType.EXPRESSION,
                                            CibRuleInEffectStatus.UNKNOWN,
                                            {
                                                "attribute": "attr2",
                                                "operation": "gt",
                                                "type": "integer",
                                                "value": "5",
                                            },
                                            None,
                                            None,
                                            [],
                                            "attr2 gt integer 5",
                                        ),
                                        CibRuleExpressionDto(
                                            f"{self.tag}-meta_attributes-rule-rule-expr-2",
                                            CibRuleExpressionType.DATE_EXPRESSION,
                                            CibRuleInEffectStatus.UNKNOWN,
                                            {
                                                "operation": "lt",
                                                "end": "2020-08-07",
                                            },
                                            None,
                                            None,
                                            [],
                                            "date lt 2020-08-07",
                                        ),
                                        CibRuleExpressionDto(
                                            f"{self.tag}-meta_attributes-rule-rule-expr-3",
                                            CibRuleExpressionType.DATE_EXPRESSION,
                                            CibRuleInEffectStatus.UNKNOWN,
                                            {
                                                "operation": "in_range",
                                                "start": "2020-09-01",
                                                "end": "2020-09-11",
                                            },
                                            None,
                                            None,
                                            [],
                                            "date in_range 2020-09-01 to 2020-09-11",
                                        ),
                                        CibRuleExpressionDto(
                                            f"{self.tag}-meta_attributes-rule-rule-expr-4",
                                            CibRuleExpressionType.DATE_EXPRESSION,
                                            CibRuleInEffectStatus.UNKNOWN,
                                            {
                                                "operation": "in_range",
                                                "start": "2020-10-01",
                                            },
                                            None,
                                            CibRuleDateCommonDto(
                                                f"{self.tag}-meta_attributes-rule-rule-expr-4-duration",
                                                {"months": "1"},
                                            ),
                                            [],
                                            "date in_range 2020-10-01 to duration months=1",
                                        ),
                                        CibRuleExpressionDto(
                                            f"{self.tag}-meta_attributes-rule-rule-expr-5",
                                            CibRuleExpressionType.DATE_EXPRESSION,
                                            CibRuleInEffectStatus.UNKNOWN,
                                            {"operation": "date_spec"},
                                            CibRuleDateCommonDto(
                                                f"{self.tag}-meta_attributes-rule-rule-expr-5-datespec",
                                                {"years": "2021-2022"},
                                            ),
                                            None,
                                            [],
                                            "date-spec years=2021-2022",
                                        ),
                                        CibRuleExpressionDto(
                                            f"{self.tag}-meta_attributes-rule-rule-expr-6",
                                            CibRuleExpressionType.DATE_EXPRESSION,
                                            CibRuleInEffectStatus.UNKNOWN,
                                            {
                                                "operation": "in_range",
                                                "end": "2020-12-11",
                                            },
                                            None,
                                            None,
                                            [],
                                            "date in_range to 2020-12-11",
                                        ),
                                    ],
                                    "defined attr1 or attr2 gt integer 5 or "
                                    "date lt 2020-08-07 or "
                                    "date in_range 2020-09-01 to 2020-09-11 or "
                                    "date in_range 2020-10-01 to duration months=1 "
                                    "or date-spec years=2021-2022 or "
                                    "date in_range to 2020-12-11",
                                ),
                            ],
                            "resource ocf:pacemaker:Dummy and "
                            "op monitor interval=30 and "
                            "(defined attr1 or attr2 gt integer 5 or "
                            "date lt 2020-08-07 or "
                            "date in_range 2020-09-01 to 2020-09-11 or "
                            "date in_range 2020-10-01 to duration months=1 or "
                            "date-spec years=2021-2022 or "
                            "date in_range to 2020-12-11)",
                        ),
                        [
                            CibNvpairDto("my-id-pair1", "name1", "value1"),
                            CibNvpairDto("my-id-pair2", "name2", "value2"),
                        ],
                    ),
                    CibNvsetDto(
                        "meta-plain",
                        {"score": "123"},
                        None,
                        [CibNvpairDto("my-id-pair3", "name1", "value1")],
                    ),
                ],
            ),
            self.command(self.env_assist.get_env(), True),
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.RULE_IN_EFFECT_STATUS_DETECTION_NOT_SUPPORTED
                ),
            ]
        )

    def _setup_rule_in_effect(self, crm_rule_check=True, crm_rule_present=True):
        defaults_xml = f"""
            <{self.tag}>
                <meta_attributes id="my-id">
                    <rule id="my-id-rule" boolean-op="and">
                        <expression
                            id="my-id-rule-expr"
                            operation="defined" attribute="attr1"
                        />
                    </rule>
                    <nvpair id="my-id-pair1" name="name1" value="value1" />
                </meta_attributes>
            </{self.tag}>
        """
        self.config.runner.cib.load(
            filename="cib-empty-3.4.xml", optional_in_conf=defaults_xml
        )
        if crm_rule_check:
            self.config.fs.isfile(
                (os.path.join(settings.pacemaker_binaries, "crm_rule")),
                return_value=crm_rule_present,
            )

    def test_crm_rule_missing(self):
        self._setup_rule_in_effect(crm_rule_present=False)
        self.assertEqual(
            CibDefaultsDto(
                meta_attributes=[
                    self.fixture_expired_dto(CibRuleInEffectStatus.UNKNOWN),
                ],
                instance_attributes=[],
            ),
            self.command(self.env_assist.get_env(), True),
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.RULE_IN_EFFECT_STATUS_DETECTION_NOT_SUPPORTED
                ),
            ]
        )

    def test_no_expire_check(self):
        self._setup_rule_in_effect(crm_rule_check=False)
        self.assertEqual(
            CibDefaultsDto(
                meta_attributes=[
                    self.fixture_expired_dto(CibRuleInEffectStatus.UNKNOWN)
                ],
                instance_attributes=[],
            ),
            self.command(self.env_assist.get_env(), False),
        )

    def test_expired(self):
        self._setup_rule_in_effect()
        self.config.runner.pcmk.get_rule_in_effect_status(
            "my-id-rule",
            RULE_EXPIRED_RETURNCODE,
        )
        self.assertEqual(
            CibDefaultsDto(
                meta_attributes=[
                    self.fixture_expired_dto(CibRuleInEffectStatus.EXPIRED)
                ],
                instance_attributes=[],
            ),
            self.command(self.env_assist.get_env(), True),
        )

    def test_not_yet_in_effect(self):
        self._setup_rule_in_effect()
        self.config.runner.pcmk.get_rule_in_effect_status(
            "my-id-rule",
            RULE_NOT_YET_IN_EFFECT_RETURNCODE,
        )
        self.assertEqual(
            CibDefaultsDto(
                meta_attributes=[
                    self.fixture_expired_dto(
                        CibRuleInEffectStatus.NOT_YET_IN_EFFECT
                    )
                ],
                instance_attributes=[],
            ),
            self.command(self.env_assist.get_env(), True),
        )

    def test_in_effect(self):
        self._setup_rule_in_effect()
        self.config.runner.pcmk.get_rule_in_effect_status(
            "my-id-rule",
            RULE_IN_EFFECT_RETURNCODE,
        )
        self.assertEqual(
            CibDefaultsDto(
                meta_attributes=[
                    self.fixture_expired_dto(CibRuleInEffectStatus.IN_EFFECT)
                ],
                instance_attributes=[],
            ),
            self.command(self.env_assist.get_env(), True),
        )

    def test_expired_error(self):
        self._setup_rule_in_effect()
        self.config.runner.pcmk.get_rule_in_effect_status(
            "my-id-rule",
            2,  # unexpected return code
        )
        self.assertEqual(
            CibDefaultsDto(
                meta_attributes=[
                    self.fixture_expired_dto(CibRuleInEffectStatus.UNKNOWN)
                ],
                instance_attributes=[],
            ),
            self.command(self.env_assist.get_env(), True),
        )


class ResourceDefaultsConfig(DefaultsConfigMixin, TestCase):
    command = staticmethod(cib_options.resource_defaults_config)
    tag = "rsc_defaults"


class OperationDefaultsConfig(DefaultsConfigMixin, TestCase):
    command = staticmethod(cib_options.operation_defaults_config)
    tag = "op_defaults"


class DefaultsRemoveMixin:
    command = lambda *args, **kwargs: None
    tag = ""

    def setUp(self):
        # pylint: disable=invalid-name
        self.env_assist, self.config = get_env_tools(self)

    def test_nothing_to_delete(self):
        self.command(self.env_assist.get_env(), [])

    def test_defaults_section_missing(self):
        self.config.runner.cib.load()
        self.env_assist.assert_raise_library_error(
            lambda: self.command(self.env_assist.get_env(), ["set1"])
        )
        self.env_assist.assert_reports(
            [
                fixture.report_not_found(
                    "set1",
                    context_type=self.tag,
                    expected_types=["options set"],
                ),
            ]
        )

    def test_success(self):
        self.config.runner.cib.load(
            optional_in_conf=f"""
                <{self.tag}>
                    <meta_attributes id="set1" />
                    <instance_attributes id="set2" />
                    <not_an_nvset id="set3" />
                    <meta_attributes id="set4" />
                    <instance_attributes id="set5" />
                </{self.tag}>
            """,
        )
        self.config.env.push_cib(
            optional_in_conf=f"""
                <{self.tag}>
                    <meta_attributes id="set1" />
                    <not_an_nvset id="set3" />
                    <meta_attributes id="set4" />
                </{self.tag}>
        """
        )
        self.command(self.env_assist.get_env(), ["set2", "set5"])

    def test_delete_all_keep_the_section(self):
        self.config.runner.cib.load(
            optional_in_conf=f"""
                <{self.tag}>
                    <meta_attributes id="set1" />
                </{self.tag}>
            """,
        )
        self.config.env.push_cib(optional_in_conf=f"<{self.tag} />")
        self.command(self.env_assist.get_env(), ["set1"])

    def test_nvset_not_found(self):
        self.config.runner.cib.load(
            optional_in_conf=f"""
                <{self.tag}>
                    <meta_attributes id="set1" />
                    <instance_attributes id="set2" />
                    <not_an_nvset id="set3" />
                    <meta_attributes id="set4" />
                    <instance_attributes id="set5" />
                </{self.tag}>
            """,
        )
        self.env_assist.assert_raise_library_error(
            lambda: self.command(
                self.env_assist.get_env(), ["set2", "set3", "setX"]
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.report_unexpected_element(
                    "set3", "not_an_nvset", ["options set"]
                ),
                fixture.report_not_found(
                    "setX",
                    context_type=self.tag,
                    expected_types=["options set"],
                ),
            ]
        )


class ResourceDefaultsRemove(DefaultsRemoveMixin, TestCase):
    command = staticmethod(cib_options.resource_defaults_remove)
    tag = "rsc_defaults"


class OperationDefaultsRemove(DefaultsRemoveMixin, TestCase):
    command = staticmethod(cib_options.operation_defaults_remove)
    tag = "op_defaults"


class DefaultsUpdateLegacyMixin:
    # This class tests legacy use cases of not providing an nvset ID
    command = lambda *args, **kwargs: None
    tag = ""
    command_for_report = None

    def setUp(self):
        # pylint: disable=invalid-name
        self.env_assist, self.config = get_env_tools(self)
        self.reports = [fixture.warn(reports.codes.DEFAULTS_CAN_BE_OVERRIDDEN)]

    def tearDown(self):
        # pylint: disable=invalid-name
        self.env_assist.assert_reports(self.reports)

    def fixture_initial_defaults(self):
        return f"""
            <{self.tag}>
                <meta_attributes id="{self.tag}-options">
                    <nvpair id="{self.tag}-options-a" name="a" value="b"/>
                    <nvpair id="{self.tag}-options-b" name="b" value="c"/>
                </meta_attributes>
            </{self.tag}>
        """

    def test_change(self):
        self.config.runner.cib.load(
            optional_in_conf=self.fixture_initial_defaults()
        )
        self.config.env.push_cib(
            optional_in_conf=f"""
            <{self.tag}>
                <meta_attributes id="{self.tag}-options">
                    <nvpair id="{self.tag}-options-a" name="a" value="B"/>
                    <nvpair id="{self.tag}-options-b" name="b" value="C"/>
                </meta_attributes>
            </{self.tag}>
        """
        )
        self.command(self.env_assist.get_env(), None, {"a": "B", "b": "C"})

    def test_add(self):
        self.config.runner.cib.load(
            optional_in_conf=self.fixture_initial_defaults()
        )
        self.config.env.push_cib(
            optional_in_conf=f"""
            <{self.tag}>
                <meta_attributes id="{self.tag}-options">
                    <nvpair id="{self.tag}-options-a" name="a" value="b"/>
                    <nvpair id="{self.tag}-options-b" name="b" value="c"/>
                    <nvpair id="{self.tag}-options-c" name="c" value="d"/>
                </meta_attributes>
            </{self.tag}>
        """
        )
        self.command(self.env_assist.get_env(), None, {"c": "d"})

    def test_remove(self):
        self.config.runner.cib.load(
            optional_in_conf=self.fixture_initial_defaults()
        )
        self.config.env.push_cib(
            remove=(
                f"./configuration/{self.tag}/meta_attributes/nvpair[@name='a']"
            )
        )
        self.command(self.env_assist.get_env(), None, {"a": ""})

    def test_add_section_if_missing(self):
        self.config.runner.cib.load()
        self.config.env.push_cib(
            optional_in_conf=f"""
            <{self.tag}>
                <meta_attributes id="{self.tag}-meta_attributes">
                    <nvpair id="{self.tag}-meta_attributes-a" name="a" value="A"/>
                </meta_attributes>
            </{self.tag}>
        """
        )
        self.command(self.env_assist.get_env(), None, {"a": "A"})

    def test_add_meta_if_missing(self):
        self.config.runner.cib.load(optional_in_conf=f"<{self.tag} />")
        self.config.env.push_cib(
            optional_in_conf=f"""
            <{self.tag}>
                <meta_attributes id="{self.tag}-meta_attributes">
                    <nvpair id="{self.tag}-meta_attributes-a" name="a" value="A"/>
                </meta_attributes>
            </{self.tag}>
        """
        )
        self.command(self.env_assist.get_env(), None, {"a": "A"})

    def test_dont_add_section_if_only_removing(self):
        self.config.runner.cib.load()
        self.command(self.env_assist.get_env(), None, {"a": "", "b": ""})

    def test_dont_add_meta_if_only_removing(self):
        self.config.runner.cib.load(optional_in_conf=f"<{self.tag} />")
        self.command(self.env_assist.get_env(), None, {"a": "", "b": ""})

    def test_keep_section_when_empty(self):
        self.config.runner.cib.load(
            optional_in_conf=self.fixture_initial_defaults()
        )
        self.config.env.push_cib(remove=f"./configuration/{self.tag}//nvpair")
        self.command(self.env_assist.get_env(), None, {"a": "", "b": ""})

    def test_ambiguous(self):
        self.config.runner.cib.load(
            optional_in_conf=f"""
                <{self.tag}>
                    <meta_attributes id="{self.tag}-options">
                        <nvpair id="{self.tag}-options-a" name="a" value="b"/>
                        <nvpair id="{self.tag}-options-b" name="b" value="c"/>
                    </meta_attributes>
                    <meta_attributes id="{self.tag}-options-1">
                        <nvpair id="{self.tag}-options-c" name="c" value="d"/>
                    </meta_attributes>
                </{self.tag}>
            """
        )
        self.env_assist.assert_raise_library_error(
            lambda: self.command(self.env_assist.get_env(), None, {"x": "y"})
        )
        self.reports = [
            fixture.error(
                reports.codes.CIB_NVSET_AMBIGUOUS_PROVIDE_NVSET_ID,
                pcs_command=self.command_for_report,
            )
        ]


class DefaultsUpdateMixin:
    command = lambda *args, **kwargs: None
    tag = ""

    def setUp(self):
        # pylint: disable=invalid-name
        self.env_assist, self.config = get_env_tools(self)

    def fixture_initial_defaults(self):
        return f"""
            <{self.tag}>
                <meta_attributes id="{self.tag}-options">
                    <nvpair id="{self.tag}-options-a" name="a" value="b"/>
                    <nvpair id="{self.tag}-options-b" name="b" value="c"/>
                    <nvpair id="{self.tag}-options-c" name="c" value="d"/>
                </meta_attributes>
            </{self.tag}>
        """

    def test_success(self):
        self.config.runner.cib.load(
            optional_in_conf=self.fixture_initial_defaults()
        )
        self.config.env.push_cib(
            optional_in_conf=f"""
            <{self.tag}>
                <meta_attributes id="{self.tag}-options">
                    <nvpair id="{self.tag}-options-a" name="a" value="B"/>
                    <nvpair id="{self.tag}-options-b" name="b" value="c"/>
                    <nvpair id="{self.tag}-options-d" name="d" value="e"/>
                </meta_attributes>
            </{self.tag}>
        """
        )
        self.command(
            self.env_assist.get_env(),
            f"{self.tag}-options",
            {"a": "B", "c": "", "d": "e"},
        )
        self.env_assist.assert_reports(
            [fixture.warn(reports.codes.DEFAULTS_CAN_BE_OVERRIDDEN)]
        )

    def test_nvset_doesnt_exist(self):
        self.config.runner.cib.load(
            optional_in_conf=self.fixture_initial_defaults()
        )
        self.env_assist.assert_raise_library_error(
            lambda: self.command(
                self.env_assist.get_env(),
                "wrong-nvset-id",
                {},
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.report_not_found(
                    "wrong-nvset-id",
                    context_type=self.tag,
                    expected_types=["options set"],
                ),
            ]
        )

    def test_keep_elements_when_empty(self):
        self.config.runner.cib.load(
            optional_in_conf=self.fixture_initial_defaults()
        )
        self.config.env.push_cib(remove=f"./configuration/{self.tag}//nvpair")
        self.command(
            self.env_assist.get_env(),
            f"{self.tag}-options",
            {"a": "", "b": "", "c": ""},
        )
        self.env_assist.assert_reports(
            [fixture.warn(reports.codes.DEFAULTS_CAN_BE_OVERRIDDEN)]
        )


class ResourceDefaultsUpdateLegacy(DefaultsUpdateLegacyMixin, TestCase):
    command = staticmethod(cib_options.resource_defaults_update)
    tag = "rsc_defaults"
    command_for_report = reports.const.PCS_COMMAND_RESOURCE_DEFAULTS_UPDATE


class OperationDefaultsUpdateLegacy(DefaultsUpdateLegacyMixin, TestCase):
    command = staticmethod(cib_options.operation_defaults_update)
    tag = "op_defaults"
    command_for_report = reports.const.PCS_COMMAND_OPERATION_DEFAULTS_UPDATE


class ResourceDefaultsUpdate(DefaultsUpdateMixin, TestCase):
    command = staticmethod(cib_options.resource_defaults_update)
    tag = "rsc_defaults"


class OperationDefaultsUpdate(DefaultsUpdateMixin, TestCase):
    command = staticmethod(cib_options.operation_defaults_update)
    tag = "op_defaults"



import base64
import os.path
from textwrap import dedent
from unittest import (
    TestCase,
    mock,
)

import pcs.lib.corosync.qdevice_net as lib
from pcs import settings
from pcs.common import reports
from pcs.common.reports import ReportItemSeverity as severity
from pcs.common.reports import codes as report_codes
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner

from pcs_test.tools.assertions import assert_raise_library_error
from pcs_test.tools.misc import (
    get_test_resource,
    get_tmp_dir,
)

# pylint: disable=no-self-use

_qnetd_cert_tool = os.path.join(
    settings.corosync_qnet_binaries, "corosync-qnetd-certutil"
)
_qnetd_tool = os.path.join(
    settings.corosync_qnet_binaries, "corosync-qnetd-tool"
)
_client_cert_tool = os.path.join(
    settings.corosync_qdevice_binaries, "corosync-qdevice-net-certutil"
)


def cert_to_url(cert):
    return base64.b64encode(cert).decode("utf-8").replace("=", "%3D")


class CertificateTestCase(TestCase):
    def setUp(self):
        self.mock_runner = mock.MagicMock(spec_set=CommandRunner)
        self.mock_tmpfile = mock.MagicMock()
        self.mock_tmpfile.name = "tmpfile path"


@mock.patch("pcs.lib.corosync.qdevice_net.qdevice_initialized")
class QdeviceSetupTest(TestCase):
    def setUp(self):
        self.mock_runner = mock.MagicMock(spec_set=CommandRunner)

    def test_success(self, mock_initialized):
        mock_initialized.return_value = False
        self.mock_runner.run.return_value = ("initialized", "", 0)

        lib.qdevice_setup(self.mock_runner)

        mock_initialized.assert_called_once_with()
        self.mock_runner.run.assert_called_once_with([_qnetd_cert_tool, "-i"])

    def test_cert_db_exists(self, mock_initialized):
        mock_initialized.return_value = True

        assert_raise_library_error(
            lambda: lib.qdevice_setup(self.mock_runner),
            (
                severity.ERROR,
                report_codes.QDEVICE_ALREADY_INITIALIZED,
                {"model": "net"},
            ),
        )

        mock_initialized.assert_called_once_with()
        self.mock_runner.run.assert_not_called()

    def test_init_tool_fail(self, mock_initialized):
        mock_initialized.return_value = False
        self.mock_runner.run.return_value = ("stdout", "test error", 1)

        assert_raise_library_error(
            lambda: lib.qdevice_setup(self.mock_runner),
            (
                severity.ERROR,
                report_codes.QDEVICE_INITIALIZATION_ERROR,
                {
                    "model": "net",
                    "reason": "test error\nstdout",
                },
            ),
        )

        mock_initialized.assert_called_once_with()
        self.mock_runner.run.assert_called_once_with([_qnetd_cert_tool, "-i"])


@mock.patch("pcs.lib.corosync.qdevice_net.shutil.rmtree")
@mock.patch("pcs.lib.corosync.qdevice_net.qdevice_initialized")
class QdeviceDestroyTest(TestCase):
    def test_success(self, mock_initialized, mock_rmtree):
        mock_initialized.return_value = True
        lib.qdevice_destroy()
        mock_rmtree.assert_called_once_with(
            settings.corosync_qdevice_net_server_certs_dir
        )

    def test_not_initialized(self, mock_initialized, mock_rmtree):
        mock_initialized.return_value = False
        lib.qdevice_destroy()
        mock_rmtree.assert_not_called()

    def test_cert_dir_rm_error(self, mock_initialized, mock_rmtree):
        mock_initialized.return_value = True
        mock_rmtree.side_effect = EnvironmentError("test errno", "test message")
        assert_raise_library_error(
            lib.qdevice_destroy,
            (
                severity.ERROR,
                report_codes.QDEVICE_DESTROY_ERROR,
                {
                    "model": "net",
                    "reason": "test message",
                },
            ),
        )
        mock_rmtree.assert_called_once_with(
            settings.corosync_qdevice_net_server_certs_dir
        )


class QdeviceStatusGenericTest(TestCase):
    def setUp(self):
        self.mock_runner = mock.MagicMock(spec_set=CommandRunner)

    def test_success(self):
        self.mock_runner.run.return_value = ("status info", "", 0)
        self.assertEqual(
            "status info", lib.qdevice_status_generic_text(self.mock_runner)
        )
        self.mock_runner.run.assert_called_once_with([_qnetd_tool, "-s"])

    def test_success_verbose(self):
        self.mock_runner.run.return_value = ("status info", "", 0)
        self.assertEqual(
            "status info",
            lib.qdevice_status_generic_text(self.mock_runner, True),
        )
        self.mock_runner.run.assert_called_once_with([_qnetd_tool, "-s", "-v"])

    def test_error(self):
        self.mock_runner.run.return_value = ("some info", "status error", 1)
        assert_raise_library_error(
            lambda: lib.qdevice_status_generic_text(self.mock_runner),
            (
                severity.ERROR,
                report_codes.QDEVICE_GET_STATUS_ERROR,
                {
                    "model": "net",
                    "reason": "status error\nsome info",
                },
            ),
        )
        self.mock_runner.run.assert_called_once_with([_qnetd_tool, "-s"])


class QdeviceStatusClusterTest(TestCase):
    def setUp(self):
        self.mock_runner = mock.MagicMock(spec_set=CommandRunner)

    def test_success(self):
        self.mock_runner.run.return_value = ("status info", "", 0)
        self.assertEqual(
            "status info", lib.qdevice_status_cluster_text(self.mock_runner)
        )
        self.mock_runner.run.assert_called_once_with([_qnetd_tool, "-l"])

    def test_success_verbose(self):
        self.mock_runner.run.return_value = ("status info", "", 0)
        self.assertEqual(
            "status info",
            lib.qdevice_status_cluster_text(self.mock_runner, verbose=True),
        )
        self.mock_runner.run.assert_called_once_with([_qnetd_tool, "-l", "-v"])

    def test_success_cluster(self):
        self.mock_runner.run.return_value = ("status info", "", 0)
        self.assertEqual(
            "status info",
            lib.qdevice_status_cluster_text(self.mock_runner, "cluster"),
        )
        self.mock_runner.run.assert_called_once_with(
            [_qnetd_tool, "-l", "-c", "cluster"]
        )

    def test_success_cluster_verbose(self):
        self.mock_runner.run.return_value = ("status info", "", 0)
        self.assertEqual(
            "status info",
            lib.qdevice_status_cluster_text(self.mock_runner, "cluster", True),
        )
        self.mock_runner.run.assert_called_once_with(
            [_qnetd_tool, "-l", "-v", "-c", "cluster"]
        )

    def test_error(self):
        self.mock_runner.run.return_value = ("some info", "status error", 1)
        assert_raise_library_error(
            lambda: lib.qdevice_status_cluster_text(self.mock_runner),
            (
                severity.ERROR,
                report_codes.QDEVICE_GET_STATUS_ERROR,
                {
                    "model": "net",
                    "reason": "status error\nsome info",
                },
            ),
        )
        self.mock_runner.run.assert_called_once_with([_qnetd_tool, "-l"])


class QdeviceConnectedClustersTest(TestCase):
    def test_empty_status(self):
        status = ""
        self.assertEqual([], lib.qdevice_connected_clusters(status))

    def test_one_cluster(self):
        status = dedent(
            """\
            Cluster "rhel72":
                Algorithm:          LMS
                Tie-breaker:        Node with lowest node ID
                Node ID 2:
                    Client address:         ::ffff:192.168.122.122:59738
                    Configured node list:   1, 2
                    Membership node list:   1, 2
                    Vote:                   ACK (ACK)
                Node ID 1:
                    Client address:         ::ffff:192.168.122.121:43420
                    Configured node list:   1, 2
                    Membership node list:   1, 2
                    Vote:                   ACK (ACK)
            """
        )
        self.assertEqual(["rhel72"], lib.qdevice_connected_clusters(status))

    def test_more_clusters(self):
        status = dedent(
            """\
            Cluster "rhel72":
            Cluster "rhel73":
            """
        )
        self.assertEqual(
            ["rhel72", "rhel73"], lib.qdevice_connected_clusters(status)
        )

    def test_invalid_status(self):
        status = dedent(
            """\
            Cluster:
                Cluster "rhel72":
            """
        )
        self.assertEqual([], lib.qdevice_connected_clusters(status))


@mock.patch("pcs.lib.corosync.qdevice_net._get_output_certificate")
@mock.patch("pcs.lib.corosync.qdevice_net._store_to_tmpfile")
class QdeviceSignCertificateRequestTest(CertificateTestCase):
    @mock.patch(
        "pcs.lib.corosync.qdevice_net.qdevice_initialized", lambda: True
    )
    def test_success(self, mock_tmp_store, mock_get_cert):
        mock_tmp_store.return_value = self.mock_tmpfile
        self.mock_runner.run.return_value = ("tool output", "", 0)
        mock_get_cert.return_value = "new certificate".encode("utf-8")

        result = lib.qdevice_sign_certificate_request(
            self.mock_runner, "certificate request", "clusterName"
        )
        self.assertEqual(result, mock_get_cert.return_value)

        mock_tmp_store.assert_called_once_with(
            "certificate request",
            reports.messages.QdeviceCertificateSignError,
        )
        self.mock_runner.run.assert_called_once_with(
            [
                _qnetd_cert_tool,
                "-s",
                "-c",
                self.mock_tmpfile.name,
                "-n",
                "clusterName",
            ]
        )
        self.assertEqual("tool output", mock_get_cert.call_args[0][0])
        self.assertEqual(
            reports.messages.QdeviceCertificateSignError("tool output"),
            mock_get_cert.call_args[0][1]("tool output"),
        )

    @mock.patch(
        "pcs.lib.corosync.qdevice_net.qdevice_initialized", lambda: False
    )
    def test_not_initialized(self, mock_tmp_store, mock_get_cert):
        assert_raise_library_error(
            lambda: lib.qdevice_sign_certificate_request(
                self.mock_runner, "certificate request", "clusterName"
            ),
            (
                severity.ERROR,
                report_codes.QDEVICE_NOT_INITIALIZED,
                {
                    "model": "net",
                },
            ),
        )
        mock_tmp_store.assert_not_called()
        self.mock_runner.run.assert_not_called()
        mock_get_cert.assert_not_called()

    @mock.patch(
        "pcs.lib.corosync.qdevice_net.qdevice_initialized", lambda: True
    )
    def test_input_write_error(self, mock_tmp_store, mock_get_cert):
        mock_tmp_store.side_effect = LibraryError

        self.assertRaises(
            LibraryError,
            lambda: lib.qdevice_sign_certificate_request(
                self.mock_runner, "certificate request", "clusterName"
            ),
        )

        self.mock_runner.run.assert_not_called()
        mock_get_cert.assert_not_called()

    @mock.patch(
        "pcs.lib.corosync.qdevice_net.qdevice_initialized", lambda: True
    )
    def test_sign_error(self, mock_tmp_store, mock_get_cert):
        mock_tmp_store.return_value = self.mock_tmpfile
        self.mock_runner.run.return_value = ("stdout", "tool output error", 1)

        assert_raise_library_error(
            lambda: lib.qdevice_sign_certificate_request(
                self.mock_runner, "certificate request", "clusterName"
            ),
            (
                severity.ERROR,
                report_codes.QDEVICE_CERTIFICATE_SIGN_ERROR,
                {
                    "reason": "tool output error\nstdout",
                },
            ),
        )

        mock_tmp_store.assert_called_once_with(
            "certificate request",
            reports.messages.QdeviceCertificateSignError,
        )
        self.mock_runner.run.assert_called_once_with(
            [
                _qnetd_cert_tool,
                "-s",
                "-c",
                self.mock_tmpfile.name,
                "-n",
                "clusterName",
            ]
        )
        mock_get_cert.assert_not_called()

    @mock.patch(
        "pcs.lib.corosync.qdevice_net.qdevice_initialized", lambda: True
    )
    def test_output_read_error(self, mock_tmp_store, mock_get_cert):
        mock_tmp_store.return_value = self.mock_tmpfile
        self.mock_runner.run.return_value = ("tool output", "", 0)
        mock_get_cert.side_effect = LibraryError

        self.assertRaises(
            LibraryError,
            lambda: lib.qdevice_sign_certificate_request(
                self.mock_runner, "certificate request", "clusterName"
            ),
        )

        mock_tmp_store.assert_called_once_with(
            "certificate request",
            reports.messages.QdeviceCertificateSignError,
        )
        self.mock_runner.run.assert_called_once_with(
            [
                _qnetd_cert_tool,
                "-s",
                "-c",
                self.mock_tmpfile.name,
                "-n",
                "clusterName",
            ]
        )
        self.assertEqual("tool output", mock_get_cert.call_args[0][0])
        self.assertEqual(
            reports.messages.QdeviceCertificateSignError("tool output"),
            mock_get_cert.call_args[0][1]("tool output"),
        )


@mock.patch("pcs.lib.corosync.qdevice_net.shutil.rmtree")
@mock.patch("pcs.lib.corosync.qdevice_net.client_initialized")
class ClientDestroyTest(TestCase):
    def test_success(self, mock_initialized, mock_rmtree):
        mock_initialized.return_value = True
        lib.client_destroy()
        mock_rmtree.assert_called_once_with(
            settings.corosync_qdevice_net_client_certs_dir
        )

    def test_not_initialized(self, mock_initialized, mock_rmtree):
        mock_initialized.return_value = False
        lib.client_destroy()
        mock_rmtree.assert_not_called()

    def test_cert_dir_rm_error(self, mock_initialized, mock_rmtree):
        mock_initialized.return_value = True
        mock_rmtree.side_effect = EnvironmentError("test errno", "test message")
        assert_raise_library_error(
            lib.client_destroy,
            (
                severity.ERROR,
                report_codes.QDEVICE_DESTROY_ERROR,
                {
                    "model": "net",
                    "reason": "test message",
                },
            ),
        )
        mock_rmtree.assert_called_once_with(
            settings.corosync_qdevice_net_client_certs_dir
        )


class ClientSetupTest(TestCase):
    def setUp(self):
        self.mock_runner = mock.MagicMock(spec_set=CommandRunner)
        self.original_path = settings.corosync_qdevice_net_client_certs_dir
        self.cert_dir = get_tmp_dir("tier0_lib_corosync_qdevice_net")
        settings.corosync_qdevice_net_client_certs_dir = self.cert_dir.name
        self.ca_file_path = os.path.join(
            settings.corosync_qdevice_net_client_certs_dir,
            settings.corosync_qdevice_net_client_ca_file_name,
        )
        with open(self.ca_file_path, "w") as my_file:
            my_file.write("to be overwritten")

    def tearDown(self):
        settings.corosync_qdevice_net_client_certs_dir = self.original_path
        self.cert_dir.cleanup()

    @mock.patch("pcs.lib.corosync.qdevice_net.client_destroy")
    def test_success(self, mock_destroy):
        self.mock_runner.run.return_value = ("tool output", "", 0)

        lib.client_setup(
            self.mock_runner, "qnetd CA certificate".encode("utf-8")
        )
        with open(self.ca_file_path, "rb") as a_file:
            self.assertEqual(
                "qnetd CA certificate".encode("utf-8"), a_file.read()
            )
        self.mock_runner.run.assert_called_once_with(
            [_client_cert_tool, "-i", "-c", self.ca_file_path]
        )
        mock_destroy.assert_called_once_with()

    @mock.patch("pcs.lib.corosync.qdevice_net.client_destroy")
    def test_init_error(self, mock_destroy):
        self.mock_runner.run.return_value = ("stdout", "tool output error", 1)

        assert_raise_library_error(
            lambda: lib.client_setup(
                self.mock_runner, "qnetd CA certificate".encode("utf-8")
            ),
            (
                severity.ERROR,
                report_codes.QDEVICE_INITIALIZATION_ERROR,
                {
                    "model": "net",
                    "reason": "tool output error\nstdout",
                },
            ),
        )
        with open(self.ca_file_path, "rb") as a_file:
            file_content = a_file.read()

        self.assertEqual("qnetd CA certificate".encode("utf-8"), file_content)
        self.mock_runner.run.assert_called_once_with(
            [_client_cert_tool, "-i", "-c", self.ca_file_path]
        )
        mock_destroy.assert_called_once_with()


@mock.patch("pcs.lib.corosync.qdevice_net._get_output_certificate")
class ClientGenerateCertificateRequestTest(CertificateTestCase):
    @mock.patch("pcs.lib.corosync.qdevice_net.client_initialized", lambda: True)
    def test_success(self, mock_get_cert):
        self.mock_runner.run.return_value = ("tool output", "", 0)
        mock_get_cert.return_value = "new certificate".encode("utf-8")

        result = lib.client_generate_certificate_request(
            self.mock_runner, "clusterName"
        )
        self.assertEqual(result, mock_get_cert.return_value)

        self.mock_runner.run.assert_called_once_with(
            [_client_cert_tool, "-r", "-n", "clusterName"]
        )
        self.assertEqual(1, len(mock_get_cert.mock_calls))
        self.assertEqual(
            "tool output",
            mock_get_cert.call_args[0][0],
        )
        self.assertEqual(
            reports.messages.QdeviceInitializationError("net", "tool output"),
            mock_get_cert.call_args[0][1]("tool output"),
        )

    @mock.patch(
        "pcs.lib.corosync.qdevice_net.client_initialized", lambda: False
    )
    def test_not_initialized(self, mock_get_cert):
        assert_raise_library_error(
            lambda: lib.client_generate_certificate_request(
                self.mock_runner, "clusterName"
            ),
            (
                severity.ERROR,
                report_codes.QDEVICE_NOT_INITIALIZED,
                {
                    "model": "net",
                },
            ),
        )
        self.mock_runner.run.assert_not_called()
        mock_get_cert.assert_not_called()

    @mock.patch("pcs.lib.corosync.qdevice_net.client_initialized", lambda: True)
    def test_tool_error(self, mock_get_cert):
        self.mock_runner.run.return_value = ("stdout", "tool output error", 1)

        assert_raise_library_error(
            lambda: lib.client_generate_certificate_request(
                self.mock_runner, "clusterName"
            ),
            (
                severity.ERROR,
                report_codes.QDEVICE_INITIALIZATION_ERROR,
                {
                    "model": "net",
                    "reason": "tool output error\nstdout",
                },
            ),
        )
        self.mock_runner.run.assert_called_once_with(
            [_client_cert_tool, "-r", "-n", "clusterName"]
        )
        mock_get_cert.assert_not_called()


@mock.patch("pcs.lib.corosync.qdevice_net._get_output_certificate")
@mock.patch("pcs.lib.corosync.qdevice_net._store_to_tmpfile")
class ClientCertRequestToPk12Test(CertificateTestCase):
    @mock.patch("pcs.lib.corosync.qdevice_net.client_initialized", lambda: True)
    def test_success(self, mock_tmp_store, mock_get_cert):
        mock_tmp_store.return_value = self.mock_tmpfile
        self.mock_runner.run.return_value = ("tool output", "", 0)
        mock_get_cert.return_value = "new certificate".encode("utf-8")

        result = lib.client_cert_request_to_pk12(
            self.mock_runner, "certificate request"
        )
        self.assertEqual(result, mock_get_cert.return_value)

        mock_tmp_store.assert_called_once_with(
            "certificate request",
            reports.messages.QdeviceCertificateImportError,
        )
        self.mock_runner.run.assert_called_once_with(
            [_client_cert_tool, "-M", "-c", self.mock_tmpfile.name]
        )
        self.assertEqual("tool output", mock_get_cert.call_args[0][0])
        self.assertEqual(
            reports.messages.QdeviceCertificateImportError("tool output"),
            mock_get_cert.call_args[0][1]("tool output"),
        )

    @mock.patch(
        "pcs.lib.corosync.qdevice_net.client_initialized", lambda: False
    )
    def test_not_initialized(self, mock_tmp_store, mock_get_cert):
        assert_raise_library_error(
            lambda: lib.client_cert_request_to_pk12(
                self.mock_runner, "certificate request"
            ),
            (
                severity.ERROR,
                report_codes.QDEVICE_NOT_INITIALIZED,
                {
                    "model": "net",
                },
            ),
        )
        mock_tmp_store.assert_not_called()
        self.mock_runner.run.assert_not_called()
        mock_get_cert.assert_not_called()

    @mock.patch("pcs.lib.corosync.qdevice_net.client_initialized", lambda: True)
    def test_input_write_error(self, mock_tmp_store, mock_get_cert):
        mock_tmp_store.side_effect = LibraryError

        self.assertRaises(
            LibraryError,
            lambda: lib.client_cert_request_to_pk12(
                self.mock_runner, "certificate request"
            ),
        )

        mock_tmp_store.assert_called_once_with(
            "certificate request",
            reports.messages.QdeviceCertificateImportError,
        )
        self.mock_runner.run.assert_not_called()
        mock_get_cert.assert_not_called()

    @mock.patch("pcs.lib.corosync.qdevice_net.client_initialized", lambda: True)
    def test_transform_error(self, mock_tmp_store, mock_get_cert):
        mock_tmp_store.return_value = self.mock_tmpfile
        self.mock_runner.run.return_value = ("stdout", "tool output error", 1)

        assert_raise_library_error(
            lambda: lib.client_cert_request_to_pk12(
                self.mock_runner, "certificate request"
            ),
            (
                severity.ERROR,
                report_codes.QDEVICE_CERTIFICATE_IMPORT_ERROR,
                {
                    "reason": "tool output error\nstdout",
                },
            ),
        )

        mock_tmp_store.assert_called_once_with(
            "certificate request",
            reports.messages.QdeviceCertificateImportError,
        )
        self.mock_runner.run.assert_called_once_with(
            [_client_cert_tool, "-M", "-c", self.mock_tmpfile.name]
        )
        mock_get_cert.assert_not_called()

    @mock.patch("pcs.lib.corosync.qdevice_net.client_initialized", lambda: True)
    def test_output_read_error(self, mock_tmp_store, mock_get_cert):
        mock_tmp_store.return_value = self.mock_tmpfile
        self.mock_runner.run.return_value = ("tool output", "", 0)
        mock_get_cert.side_effect = LibraryError

        self.assertRaises(
            LibraryError,
            lambda: lib.client_cert_request_to_pk12(
                self.mock_runner, "certificate request"
            ),
        )

        mock_tmp_store.assert_called_once_with(
            "certificate request",
            reports.messages.QdeviceCertificateImportError,
        )
        self.mock_runner.run.assert_called_once_with(
            [_client_cert_tool, "-M", "-c", self.mock_tmpfile.name]
        )
        self.assertEqual("tool output", mock_get_cert.call_args[0][0])
        self.assertEqual(
            reports.messages.QdeviceCertificateImportError("tool output"),
            mock_get_cert.call_args[0][1]("tool output"),
        )


@mock.patch("pcs.lib.corosync.qdevice_net._store_to_tmpfile")
class ClientImportCertificateAndKeyTest(CertificateTestCase):
    @mock.patch("pcs.lib.corosync.qdevice_net.client_initialized", lambda: True)
    def test_success(self, mock_tmp_store):
        mock_tmp_store.return_value = self.mock_tmpfile
        self.mock_runner.run.return_value = ("tool output", "", 0)

        lib.client_import_certificate_and_key(
            self.mock_runner, "pk12 certificate"
        )

        mock_tmp_store.assert_called_once_with(
            "pk12 certificate",
            reports.messages.QdeviceCertificateImportError,
        )
        self.mock_runner.run.assert_called_once_with(
            [_client_cert_tool, "-m", "-c", self.mock_tmpfile.name]
        )

    @mock.patch(
        "pcs.lib.corosync.qdevice_net.client_initialized", lambda: False
    )
    def test_not_initialized(self, mock_tmp_store):
        assert_raise_library_error(
            lambda: lib.client_import_certificate_and_key(
                self.mock_runner, "pk12 certificate"
            ),
            (
                severity.ERROR,
                report_codes.QDEVICE_NOT_INITIALIZED,
                {
                    "model": "net",
                },
            ),
        )

        mock_tmp_store.assert_not_called()
        self.mock_runner.run.assert_not_called()

    @mock.patch("pcs.lib.corosync.qdevice_net.client_initialized", lambda: True)
    def test_input_write_error(self, mock_tmp_store):
        mock_tmp_store.side_effect = LibraryError

        self.assertRaises(
            LibraryError,
            lambda: lib.client_import_certificate_and_key(
                self.mock_runner, "pk12 certificate"
            ),
        )

        mock_tmp_store.assert_called_once_with(
            "pk12 certificate",
            reports.messages.QdeviceCertificateImportError,
        )
        self.mock_runner.run.assert_not_called()

    @mock.patch("pcs.lib.corosync.qdevice_net.client_initialized", lambda: True)
    def test_import_error(self, mock_tmp_store):
        mock_tmp_store.return_value = self.mock_tmpfile
        self.mock_runner.run.return_value = ("stdout", "tool output error", 1)

        assert_raise_library_error(
            lambda: lib.client_import_certificate_and_key(
                self.mock_runner, "pk12 certificate"
            ),
            (
                severity.ERROR,
                report_codes.QDEVICE_CERTIFICATE_IMPORT_ERROR,
                {
                    "reason": "tool output error\nstdout",
                },
            ),
        )

        mock_tmp_store.assert_called_once_with(
            "pk12 certificate",
            reports.messages.QdeviceCertificateImportError,
        )
        mock_tmp_store.assert_called_once_with(
            "pk12 certificate",
            reports.messages.QdeviceCertificateImportError,
        )
        self.mock_runner.run.assert_called_once_with(
            [_client_cert_tool, "-m", "-c", self.mock_tmpfile.name]
        )


class GetOutputCertificateTest(TestCase):
    # pylint: disable=protected-access
    def setUp(self):
        self.file_path = get_test_resource("qdevice-certs/qnetd-cacert.crt")
        with open(self.file_path, "rb") as a_file:
            self.file_data = a_file.read()

    def test_success(self):
        cert_tool_output = dedent(
            """
            some line
            Certificate stored in {0}
            some other line
            """.format(
                self.file_path
            )
        )
        report_func = mock.MagicMock()

        self.assertEqual(
            self.file_data,
            lib._get_output_certificate(cert_tool_output, report_func),
        )
        report_func.assert_not_called()

    def test_success_request(self):
        cert_tool_output = dedent(
            """
            some line
            Certificate request stored in {0}
            some other line
            """.format(
                self.file_path
            )
        )
        report_func = mock.MagicMock()

        self.assertEqual(
            self.file_data,
            lib._get_output_certificate(cert_tool_output, report_func),
        )
        report_func.assert_not_called()

    def test_message_not_found(self):
        cert_tool_output = "some rubbish output"
        report_item_message = reports.messages.QdeviceCertificateImportError

        assert_raise_library_error(
            lambda: lib._get_output_certificate(
                cert_tool_output,
                report_item_message,
            ),
            (
                severity.ERROR,
                report_codes.QDEVICE_CERTIFICATE_IMPORT_ERROR,
                {
                    "reason": cert_tool_output,
                },
            ),
        )

    def test_cannot_read_file(self):
        cert_tool_output = dedent(
            """
            some line
            Certificate request stored in {0}.bad
            some other line
            """.format(
                self.file_path
            )
        )
        report_item_message = reports.messages.QdeviceCertificateImportError

        assert_raise_library_error(
            lambda: lib._get_output_certificate(
                cert_tool_output,
                report_item_message,
            ),
            (
                severity.ERROR,
                report_codes.QDEVICE_CERTIFICATE_IMPORT_ERROR,
                {
                    "reason": "{0}.bad: No such file or directory".format(
                        self.file_path
                    ),
                },
            ),
        )



"""
Module for stuff related to clones.

Previously, promotable clones were implemented in pacemaker as 'master'
elements whereas regular clones were 'clone' elements. Since pacemaker-2.0,
promotable clones are clones with meta attribute promotable=true. Master
elements are deprecated yet still supported in pacemaker. We provide read-only
support for them to be able to read, process and display CIBs containing them.
"""
from typing import (
    List,
    Mapping,
    Optional,
    cast,
)

import dataclasses
from lxml import etree
from lxml.etree import _Element

from pcs.common.pacemaker import nvset
from pcs.common.pacemaker.resource.clone import CibResourceCloneDto
from pcs.common.reports import ReportItemList
from pcs.lib.cib import (
    nvpair,
    nvpair_multi,
    rule,
)
from pcs.lib.cib.const import TAG_RESOURCE_CLONE as TAG_CLONE
from pcs.lib.cib.const import TAG_RESOURCE_MASTER as TAG_MASTER
from pcs.lib.cib.tools import IdProvider
from pcs.lib.pacemaker.values import (
    is_true,
    validate_id,
)

ALL_TAGS = [TAG_CLONE, TAG_MASTER]


def is_clone(resource_el: _Element) -> bool:
    return resource_el.tag == TAG_CLONE


def is_master(resource_el: _Element) -> bool:
    return resource_el.tag == TAG_MASTER


def is_any_clone(resource_el: _Element) -> bool:
    return resource_el.tag in ALL_TAGS


def is_promotable_clone(resource_el: _Element) -> bool:
    """
    Return True if resource_el is a promotable clone, False on clone and master
    """
    return is_clone(resource_el) and is_true(
        nvpair.get_value(
            nvpair.META_ATTRIBUTES_TAG,
            resource_el,
            "promotable",
            default="false",
        )
    )


def clone_element_to_dto(
    clone_element: _Element,
    rule_eval: Optional[rule.RuleInEffectEval] = None,
) -> CibResourceCloneDto:
    if rule_eval is None:
        rule_eval = rule.RuleInEffectEvalDummy()
    return CibResourceCloneDto(
        id=str(clone_element.attrib["id"]),
        description=clone_element.get("description"),
        member_id=str(get_inner_resource(clone_element).attrib["id"]),
        meta_attributes=[
            nvpair_multi.nvset_element_to_dto(nvset, rule_eval)
            for nvset in nvpair_multi.find_nvsets(
                clone_element, nvpair_multi.NVSET_META
            )
        ],
        instance_attributes=[
            nvpair_multi.nvset_element_to_dto(nvset, rule_eval)
            for nvset in nvpair_multi.find_nvsets(
                clone_element, nvpair_multi.NVSET_INSTANCE
            )
        ],
    )


def master_element_to_dto(
    master_element: _Element,
    rule_eval: Optional[rule.RuleInEffectEval] = None,
) -> CibResourceCloneDto:
    clone_dto = clone_element_to_dto(master_element, rule_eval)
    promotable_nvpair = nvset.CibNvpairDto(
        id="", name="promotable", value="true"
    )
    if clone_dto.meta_attributes:
        first_meta_attributes = clone_dto.meta_attributes[0]
        clone_dto = dataclasses.replace(
            clone_dto,
            meta_attributes=[
                dataclasses.replace(
                    first_meta_attributes,
                    nvpairs=(
                        list(first_meta_attributes.nvpairs)
                        + [promotable_nvpair]
                    ),
                )
            ]
            + list(clone_dto.meta_attributes[1:]),
        )
    else:
        clone_dto = dataclasses.replace(
            clone_dto,
            meta_attributes=[
                nvset.CibNvsetDto(
                    id="", options={}, rule=None, nvpairs=[promotable_nvpair]
                )
            ],
        )

    return clone_dto


def get_parent_any_clone(resource_el: _Element) -> Optional[_Element]:
    """
    Get any parent clone of a primitive (may be in a group) or group

    resource_el -- the primitive or group to get its parent clone
    """
    element = resource_el
    for _ in range(2):
        parent_el = element.getparent()
        if parent_el is None:
            return None
        if is_any_clone(parent_el):
            return parent_el
        element = parent_el
    return None


def append_new(
    resources_section: _Element,
    id_provider: IdProvider,
    primitive_element: _Element,
    options: Mapping[str, str],
    clone_id: Optional[str] = None,
) -> _Element:
    """
    Append a new clone element (containing the primitive_element) to the
    resources_section.

    resources_section -- place where the new clone will be appended
    id_provider -- elements' ids generator
    primitive_element -- resource which will be cloned
    options -- source for clone meta options
    clone_id -- optional custom clone id
    """
    clone_element = etree.SubElement(
        resources_section,
        TAG_CLONE,
        id=id_provider.allocate_id(
            "{0}-{1}".format(str(primitive_element.get("id")), TAG_CLONE)
        )
        if clone_id is None
        else clone_id,
    )
    clone_element.append(primitive_element)

    if options:
        nvpair.append_new_meta_attributes(clone_element, options, id_provider)

    return clone_element


def get_inner_resource(clone_el: _Element) -> _Element:
    return cast(List[_Element], clone_el.xpath("./primitive | ./group"))[0]


def validate_clone_id(clone_id: str, id_provider: IdProvider) -> ReportItemList:
    """
    Validate that clone_id is a valid xml id and it is unique in the cib.

    clone_id -- identifier of clone element
    id_provider -- elements' ids generator
    """
    report_list: ReportItemList = []
    validate_id(clone_id, reporter=report_list)
    report_list.extend(id_provider.book_ids(clone_id))
    return report_list
